########################################################################
# File name: test_service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest

import aioxmpp.disco
import aioxmpp.service

import aioxmpp.ping.service as ping_service
import aioxmpp.ping.xso as ping_xso

from aioxmpp.testutils import (
    make_connected_client,
    run_coroutine,
)


TEST_PEER = aioxmpp.JID.fromstr("juliet@capulet.lit/balcony")


class TestService(unittest.TestCase):
    def setUp(self):
        self.cc = make_connected_client()
        self.disco_server = unittest.mock.Mock()
        self.s = ping_service.PingService(self.cc, dependencies={
            aioxmpp.DiscoServer: self.disco_server
        })

    def test_is_service(self):
        self.assertTrue(issubclass(
            ping_service.PingService,
            aioxmpp.service.Service,
        ))

    def test_register_ping_feature(self):
        self.assertIsInstance(
            ping_service.PingService._ping_feature,
            aioxmpp.disco.register_feature,
        )
        self.assertEqual(
            ping_service.PingService._ping_feature.feature,
            "urn:xmpp:ping",
        )

    def test_ping_sends_ping(self):
        self.cc.send.return_value = ping_xso.Ping()

        run_coroutine(self.s.ping(TEST_PEER))

        self.cc.send.assert_called_once_with(unittest.mock.ANY)

        _, (iq, ), _ = self.cc.send.mock_calls[0]

        self.assertIsInstance(
            iq,
            aioxmpp.IQ,
        )

        self.assertEqual(
            iq.to,
            TEST_PEER,
        )

        self.assertEqual(
            iq.type_,
            aioxmpp.IQType.GET,
        )

        self.assertIsInstance(
            iq.payload,
            ping_xso.Ping,
        )

    def test_ping_reraises_random_errors_from_send(self):
        class FooException(Exception):
            pass

        self.cc.send.side_effect = FooException()

        with self.assertRaises(FooException):
            run_coroutine(self.s.ping(TEST_PEER))

    def test_ping_reraises_XMPPErrors(self):
        exc = aioxmpp.errors.XMPPError(
            aioxmpp.ErrorCondition.NOT_AUTHORIZED,
        )
        exc.TYPE = unittest.mock.sentinel.type_
        self.cc.send.side_effect = exc

        with self.assertRaises(aioxmpp.errors.XMPPError) as ctx:
            run_coroutine(self.s.ping(TEST_PEER))

        self.assertIs(ctx.exception, exc)

    def test_handle_ping_is_decorated(self):
        self.assertTrue(aioxmpp.service.is_iq_handler(
            aioxmpp.IQType.GET,
            ping_xso.Ping,
            ping_service.PingService.handle_ping,
        ))

    def test_handle_ping_returns_pong(self):
        result = run_coroutine(self.s.handle_ping(
            unittest.mock.sentinel.request
        ))

        self.assertIsInstance(result, ping_xso.Ping)



########################################################################
# File name: test_tracking.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import contextlib
import itertools
import unittest
import unittest.mock

from datetime import timedelta

import aioxmpp.service
import aioxmpp.tracking as tracking

from aioxmpp.utils import namespaces

from aioxmpp.testutils import (
    make_connected_client,
)


TEST_LOCAL = aioxmpp.JID.fromstr("romeo@montague.example/garden")
TEST_PEER = aioxmpp.JID.fromstr("juliet@capulet.example/chamber")


class TestMessageTracker(unittest.TestCase):
    def setUp(self):
        self.t = tracking.MessageTracker()
        self.listener = unittest.mock.Mock()
        for ev in ["on_closed", "on_state_changed"]:
            cb = getattr(self.listener, ev)
            cb.return_value = None
            getattr(self.t, ev).connect(cb)

    def tearDown(self):
        del self.t

    def test_init(self):
        self.assertIsNone(self.t.response)
        self.assertEqual(
            self.t.state,
            tracking.MessageState.IN_TRANSIT
        )
        self.assertIs(self.t.closed, False)

    def test_state_is_not_writable(self):
        with self.assertRaises(AttributeError):
            self.t.state = self.t.state

    def test_response_is_not_writable(self):
        with self.assertRaises(AttributeError):
            self.t.response = self.t.response

    def test_closed_is_not_writable(self):
        with self.assertRaises(AttributeError):
            self.t.closed = self.t.closed

    def test_close_sets_closed_to_true(self):
        self.t.close()
        self.assertIs(self.t.closed, True)

    def test_close_fires_event(self):
        self.t.close()
        self.listener.on_closed.assert_called_once_with()
        self.listener.on_state_changed.assert_not_called()

    def test_close_is_idempotent(self):
        self.t.close()
        self.listener.on_closed.assert_called_once_with()
        self.listener.on_state_changed.assert_not_called()
        self.t.close()
        self.listener.on_closed.assert_called_once_with()
        self.listener.on_state_changed.assert_not_called()

    def test__set_state_updates_state_and_response(self):
        self.t._set_state(
            tracking.MessageState.ERROR,
            unittest.mock.sentinel.response,
        )

        self.assertEqual(
            self.t.state,
            tracking.MessageState.ERROR,
        )
        self.assertEqual(
            self.t.response,
            unittest.mock.sentinel.response,
        )

    def test__set_state_fires_event(self):
        self.t._set_state(
            tracking.MessageState.DELIVERED_TO_RECIPIENT,
            unittest.mock.sentinel.response,
        )
        self.listener.on_state_changed.assert_called_once_with(
            tracking.MessageState.DELIVERED_TO_RECIPIENT,
            unittest.mock.sentinel.response,
        )

    def test__set_state_rejects_transitions_from_aborted(self):
        self.t._set_state(
            tracking.MessageState.ABORTED,
            unittest.mock.sentinel.response,
        )
        self.listener.on_state_changed.reset_mock()
        self.listener.on_state_changed.return_value = None

        for state in tracking.MessageState:
            with self.assertRaisesRegex(
                    ValueError,
                    "transition from .* to .*not allowed"):
                self.t._set_state(state)
            self.listener.on_state_changed.assert_not_called()
            self.assertEqual(
                self.t.state,
                tracking.MessageState.ABORTED,
            )
            self.assertEqual(
                self.t.response,
                unittest.mock.sentinel.response,
            )

    def test__set_state_rejects_transitions_from_error_to_in_transit(self):
        self.t._set_state(
            tracking.MessageState.ERROR,
            unittest.mock.sentinel.response,
        )
        self.listener.on_state_changed.reset_mock()
        self.listener.on_state_changed.return_value = None

        with self.assertRaisesRegex(
                ValueError,
                "transition from .* to .*not allowed"):
            self.t._set_state(tracking.MessageState.IN_TRANSIT)

        self.listener.on_state_changed.assert_not_called()
        self.assertEqual(
            self.t.state,
            tracking.MessageState.ERROR,
        )
        self.assertEqual(
            self.t.response,
            unittest.mock.sentinel.response,
        )

    def test__set_state_rejects_transitions_from_error_to_aborted(self):
        self.t._set_state(
            tracking.MessageState.ERROR,
            unittest.mock.sentinel.response,
        )
        self.listener.on_state_changed.reset_mock()
        self.listener.on_state_changed.return_value = None

        with self.assertRaisesRegex(
                ValueError,
                "transition from .* to .*not allowed"):
            self.t._set_state(tracking.MessageState.ABORTED)

        self.listener.on_state_changed.assert_not_called()
        self.assertEqual(
            self.t.state,
            tracking.MessageState.ERROR,
        )
        self.assertEqual(
            self.t.response,
            unittest.mock.sentinel.response,
        )

    def test__set_state_rejects_transitions_from_error_to_delivered_to_server(
            self):
        self.t._set_state(
            tracking.MessageState.ERROR,
            unittest.mock.sentinel.response,
        )
        self.listener.on_state_changed.reset_mock()
        self.listener.on_state_changed.return_value = None

        with self.assertRaisesRegex(
                ValueError,
                "transition from .* to .*not allowed"):
            self.t._set_state(tracking.MessageState.DELIVERED_TO_SERVER)

        self.listener.on_state_changed.assert_not_called()
        self.assertEqual(
            self.t.state,
            tracking.MessageState.ERROR,
        )
        self.assertEqual(
            self.t.response,
            unittest.mock.sentinel.response,
        )

    def test__set_state_allows_transition_from_error_to_delivered_to_recipient(
            self):
        self.t._set_state(
            tracking.MessageState.ERROR,
            unittest.mock.sentinel.response,
        )
        self.listener.on_state_changed.reset_mock()
        self.listener.on_state_changed.return_value = None

        self.t._set_state(
            tracking.MessageState.DELIVERED_TO_RECIPIENT,
            unittest.mock.sentinel.response,
        )

        self.listener.on_state_changed.assert_called_once_with(
            tracking.MessageState.DELIVERED_TO_RECIPIENT,
            unittest.mock.sentinel.response,
        )

    def test__set_state_rejects_transitions_to_in_transit(self):
        for state in tracking.MessageState:
            t = tracking.MessageTracker()
            if state != t.state:
                t._set_state(state, unittest.mock.sentinel.response)
            on_state_changed = unittest.mock.Mock()
            t.on_state_changed.connect(on_state_changed)
            with self.assertRaisesRegex(
                    ValueError,
                    "transition from .* to .*not allowed"):
                t._set_state(tracking.MessageState.IN_TRANSIT)
            on_state_changed.assert_not_called()
            self.assertEqual(
                t.state,
                state,
            )
            if state != t.state:
                self.assertEqual(
                    t.response,
                    unittest.mock.sentinel.response,
                )

    def test__set_state_rejects_some_other_transitions(self):
        to_reject = [
            (
                tracking.MessageState.DELIVERED_TO_RECIPIENT,
                tracking.MessageState.DELIVERED_TO_SERVER,
            ),
            (
                tracking.MessageState.SEEN_BY_RECIPIENT,
                tracking.MessageState.DELIVERED_TO_SERVER,
            ),
            (
                tracking.MessageState.SEEN_BY_RECIPIENT,
                tracking.MessageState.DELIVERED_TO_RECIPIENT,
            ),
        ]

        for state1, state2 in itertools.product(
                tracking.MessageState,
                tracking.MessageState):
            if state1 == tracking.MessageState.ABORTED:
                # already tested elsewhere
                continue
            if state1 == tracking.MessageState.ERROR:
                # already tested elsewhere
                continue
            if state2 == tracking.MessageState.IN_TRANSIT:
                # already tested elsewhere
                continue
            t = tracking.MessageTracker()

            if state1 != t.state:
                t._set_state(state1, unittest.mock.sentinel.response)

            on_state_changed = unittest.mock.Mock()
            t.on_state_changed.connect(on_state_changed)

            if (state1, state2) in to_reject:
                with self.assertRaisesRegex(
                        ValueError,
                        "transition from .* to .*not allowed",
                        msg=(state1, state2)):
                    t._set_state(state2)
                on_state_changed.assert_not_called()
                self.assertEqual(
                    t.state,
                    state1,
                )
                if state1 != t.state:
                    self.assertEqual(
                        t.response,
                        unittest.mock.sentinel.response,
                    )

            else:
                t._set_state(
                    state2,
                    unittest.mock.sentinel.response2
                )
                self.assertEqual(
                    t.state,
                    state2,
                )
                self.assertEqual(
                    t.response,
                    unittest.mock.sentinel.response2,
                )

    def test__set_state_bails_out_early_if_closed(self):
        self.t._set_state(
            tracking.MessageState.DELIVERED_TO_SERVER,
            unittest.mock.sentinel.response,
        )

        self.t.close()

        for s in tracking.MessageState:
            with self.assertRaisesRegex(
                    RuntimeError,
                    "tracker is closed"):
                self.t._set_state(s)
            self.assertEqual(
                self.t.state,
                tracking.MessageState.DELIVERED_TO_SERVER,
            )
            self.assertEqual(
                self.t.response,
                unittest.mock.sentinel.response,
            )

    def test__set_timeout_with_number_calls_call_later(self):
        with contextlib.ExitStack() as stack:
            get_event_loop = stack.enter_context(unittest.mock.patch(
                "asyncio.get_event_loop",
            ))

            self.t.set_timeout(unittest.mock.sentinel.timeout)

        get_event_loop.assert_called_once_with()
        get_event_loop().call_later.assert_called_once_with(
            unittest.mock.sentinel.timeout,
            self.t.close,
        )

    def test__set_timeout_with_timedelta_calls_call_later(self):
        with contextlib.ExitStack() as stack:
            get_event_loop = stack.enter_context(unittest.mock.patch(
                "asyncio.get_event_loop",
            ))

            self.t.set_timeout(timedelta(days=1))

        get_event_loop.assert_called_once_with()
        get_event_loop().call_later.assert_called_once_with(
            timedelta(days=1).total_seconds(),
            self.t.close,
        )


class TestBasicTrackingService(unittest.TestCase):
    def setUp(self):
        self.cc = make_connected_client()
        self.s = tracking.BasicTrackingService(self.cc)

    def tearDown(self):
        del self.s
        del self.cc

    def test_is_service(self):
        self.assertTrue(issubclass(
            tracking.BasicTrackingService,
            aioxmpp.service.Service,
        ))

    def test_installs_message_filter(self):
        self.assertTrue(aioxmpp.service.is_inbound_message_filter(
            tracking.BasicTrackingService._inbound_message_filter,
        ))

    def test_inbound_message_filter_forwards_stanzas(self):
        for type_ in aioxmpp.MessageType:
            msg = aioxmpp.Message(
                type_=type_,
                from_=TEST_PEER,
                to=TEST_LOCAL,
            )
            self.assertIs(
                msg,
                self.s._inbound_message_filter(msg),
            )

    def test_inbound_message_filter_forwards_broken_stanzas(self):
        msg = object()
        self.assertIs(
            msg,
            self.s._inbound_message_filter(msg),
        )

    def test_attach_tracker_calls_autoset_id(self):
        msg = unittest.mock.Mock()
        self.s.attach_tracker(msg)
        msg.autoset_id.assert_called_once_with()

    def test_attach_tracker_with_subsequent_error_modifies_tracking(self):
        tracker = tracking.MessageTracker()
        msg = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=TEST_LOCAL,
            to=TEST_PEER,
        )
        self.assertIs(
            self.s.attach_tracker(msg, tracker),
            tracker
        )

        error = msg.make_error(aioxmpp.stanza.Error.from_exception(
            aioxmpp.XMPPCancelError(
                aioxmpp.ErrorCondition.FEATURE_NOT_IMPLEMENTED
            )
        ))

        self.assertIsNone(self.s._inbound_message_filter(error))

        self.assertEqual(
            tracker.state,
            tracking.MessageState.ERROR
        )
        self.assertIs(
            tracker.response,
            error,
        )

    def test_attach_tracker_with_subsequent_error_from_bare_modifies_tracking(
            self):
        tracker = tracking.MessageTracker()
        msg = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=TEST_LOCAL,
            to=TEST_PEER,
        )
        self.assertIs(
            self.s.attach_tracker(msg, tracker),
            tracker
        )

        error = msg.make_error(aioxmpp.stanza.Error.from_exception(
            aioxmpp.XMPPCancelError(
                aioxmpp.ErrorCondition.FEATURE_NOT_IMPLEMENTED
            )
        ))
        error.from_ = error.from_.bare()

        self.assertIsNone(self.s._inbound_message_filter(error))

        self.assertEqual(
            tracker.state,
            tracking.MessageState.ERROR
        )
        self.assertIs(
            tracker.response,
            error,
        )

    def test_attach_tracker_with_subsequent_error_from_other_id_untracked(
            self):
        tracker = tracking.MessageTracker()
        msg = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=TEST_LOCAL,
            to=TEST_PEER,
        )
        self.assertIs(
            self.s.attach_tracker(msg, tracker),
            tracker
        )

        error = msg.make_error(aioxmpp.stanza.Error.from_exception(
            aioxmpp.XMPPCancelError(
                aioxmpp.ErrorCondition.FEATURE_NOT_IMPLEMENTED
            )
        ))
        error.id_ = "fnord"

        self.assertIs(self.s._inbound_message_filter(error), error)

        self.assertEqual(
            tracker.state,
            tracking.MessageState.IN_TRANSIT
        )
        self.assertIs(
            tracker.response,
            None,
        )

    def test_attach_tracker_sets_delivered_to_server_if_ok(self):
        tracker = tracking.MessageTracker()
        msg = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=TEST_LOCAL,
            to=TEST_PEER,
        )
        token = unittest.mock.Mock()
        self.assertIs(
            self.s.attach_tracker(msg, tracker, token),
            tracker
        )

        token.future.add_done_callback.assert_called_once_with(
            unittest.mock.ANY,
        )

        _, (cb, ), _ = token.future.add_done_callback.mock_calls[0]

        cb(token.future)

        self.assertEqual(
            tracker.state,
            tracking.MessageState.DELIVERED_TO_SERVER,
        )

    def test_attach_tracker_sets_aborted_if_aborted(self):
        tracker = tracking.MessageTracker()
        msg = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=TEST_LOCAL,
            to=TEST_PEER,
        )
        token = unittest.mock.Mock()
        self.assertIs(
            self.s.attach_tracker(msg, tracker, token),
            tracker
        )

        token.future.add_done_callback.assert_called_once_with(
            unittest.mock.ANY,
        )
        token.future.result.side_effect = RuntimeError()

        _, (cb, ), _ = token.future.add_done_callback.mock_calls[0]

        cb(token.future)

        self.assertEqual(
            tracker.state,
            tracking.MessageState.ABORTED,
        )

    def test_attach_tracker_sets_aborted_on_other_exception(self):
        tracker = tracking.MessageTracker()
        msg = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=TEST_LOCAL,
            to=TEST_PEER,
        )
        token = unittest.mock.Mock()
        self.assertIs(
            self.s.attach_tracker(msg, tracker, token),
            tracker
        )

        class FooException(Exception):
            pass

        token.future.add_done_callback.assert_called_once_with(
            unittest.mock.ANY,
        )
        token.future.result.side_effect = FooException()

        _, (cb, ), _ = token.future.add_done_callback.mock_calls[0]

        cb(token.future)

        self.assertEqual(
            tracker.state,
            tracking.MessageState.ABORTED,
        )

    def test_attach_tracker_handler_does_not_raise_exception_if_state_already_set(self):
        tracker = tracking.MessageTracker()
        msg = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=TEST_LOCAL,
            to=TEST_PEER,
        )
        token = unittest.mock.Mock()
        self.assertIs(
            self.s.attach_tracker(msg, tracker, token),
            tracker
        )

        token.future.add_done_callback.assert_called_once_with(
            unittest.mock.ANY,
        )

        _, (cb, ), _ = token.future.add_done_callback.mock_calls[0]

        tracker._set_state(tracking.MessageState.DELIVERED_TO_RECIPIENT)

        cb(token.future)

        self.assertEqual(
            tracker.state,
            tracking.MessageState.DELIVERED_TO_RECIPIENT,
        )

    def test_attach_tracker_ignores_cancelled_stanza_token(self):
        tracker = tracking.MessageTracker()
        msg = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=TEST_LOCAL,
            to=TEST_PEER,
        )
        token = unittest.mock.Mock()
        self.assertIs(
            self.s.attach_tracker(msg, tracker, token),
            tracker
        )

        token.future.add_done_callback.assert_called_once_with(
            unittest.mock.ANY,
        )
        token.future.result.side_effect = asyncio.CancelledError()

        _, (cb, ), _ = token.future.add_done_callback.mock_calls[0]

        cb(token.future)

        self.assertEqual(
            tracker.state,
            tracking.MessageState.IN_TRANSIT,
        )

    def test_attach_tracker_does_not_set_error_if_in_delivered_state(self):
        tracker = tracking.MessageTracker()
        msg = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=TEST_LOCAL,
            to=TEST_PEER,
        )
        self.assertIs(
            self.s.attach_tracker(msg, tracker),
            tracker
        )

        tracker._set_state(tracking.MessageState.DELIVERED_TO_RECIPIENT)

        error = msg.make_error(aioxmpp.stanza.Error.from_exception(
            aioxmpp.XMPPCancelError(
                aioxmpp.ErrorCondition.FEATURE_NOT_IMPLEMENTED
            )
        ))

        self.assertIsNone(self.s._inbound_message_filter(error))

        self.assertEqual(
            tracker.state,
            tracking.MessageState.DELIVERED_TO_RECIPIENT
        )
        self.assertIs(
            tracker.response,
            None,
        )

    def test_attach_tracker_does_not_set_error_if_in_seen_state(self):
        tracker = tracking.MessageTracker()
        msg = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=TEST_LOCAL,
            to=TEST_PEER,
        )
        self.assertIs(
            self.s.attach_tracker(msg, tracker),
            tracker
        )

        tracker._set_state(tracking.MessageState.SEEN_BY_RECIPIENT)

        error = msg.make_error(aioxmpp.stanza.Error.from_exception(
            aioxmpp.XMPPCancelError(
                aioxmpp.ErrorCondition.FEATURE_NOT_IMPLEMENTED
            )
        ))

        self.assertIsNone(self.s._inbound_message_filter(error))

        self.assertEqual(
            tracker.state,
            tracking.MessageState.SEEN_BY_RECIPIENT
        )
        self.assertIs(
            tracker.response,
            None,
        )

    def test_attach_tracker_with_subsequent_chat_is_not_tracked(self):
        tracker = tracking.MessageTracker()
        msg = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=TEST_LOCAL,
            to=TEST_PEER,
        )
        self.assertIs(
            self.s.attach_tracker(msg, tracker),
            tracker
        )

        reply = msg.make_reply()
        reply.id_ = msg.id_
        self.assertIs(
            reply,
            self.s._inbound_message_filter(reply)
        )

        self.assertEqual(
            tracker.state,
            tracking.MessageState.IN_TRANSIT
        )

    def test_inbound_message_filter_ignores_closed_tracker(self):
        tracker = tracking.MessageTracker()
        msg = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=TEST_LOCAL,
            to=TEST_PEER,
        )
        self.assertIs(
            self.s.attach_tracker(msg, tracker),
            tracker
        )

        tracker.close()

        error = msg.make_error(aioxmpp.stanza.Error.from_exception(
            aioxmpp.XMPPCancelError(
                aioxmpp.ErrorCondition.FEATURE_NOT_IMPLEMENTED
            )
        ))

        self.assertIs(error, self.s._inbound_message_filter(error))

        self.assertEqual(
            tracker.state,
            tracking.MessageState.IN_TRANSIT,
        )
        self.assertTrue(tracker.closed)

    def test_attach_tracker_autocreates_tracker_if_needed(self):
        msg = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=TEST_LOCAL,
            to=TEST_PEER,
        )
        tracker = self.s.attach_tracker(msg, None)
        self.assertIsInstance(tracker, tracking.MessageTracker)

        error = msg.make_error(aioxmpp.stanza.Error.from_exception(
            aioxmpp.XMPPCancelError(
                aioxmpp.ErrorCondition.FEATURE_NOT_IMPLEMENTED
            )
        ))

        self.assertIsNone(self.s._inbound_message_filter(error))

        self.assertEqual(
            tracker.state,
            tracking.MessageState.ERROR
        )
        self.assertIs(
            tracker.response,
            error,
        )

    def test_send_tracked_attaches_and_returns_tracker(self):
        tracker = unittest.mock.sentinel.tracker
        msg = unittest.mock.sentinel.message

        with contextlib.ExitStack() as stack:
            attach_tracker = stack.enter_context(unittest.mock.patch.object(
                self.s,
                "attach_tracker",
            ))

            result = self.s.send_tracked(msg, tracker)

            self.cc.enqueue.assert_called_once_with(
                msg,
            )

            attach_tracker.assert_called_once_with(
                msg,
                tracker,
                self.cc.enqueue(),
            )

            self.assertEqual(
                result,
                self.cc.enqueue(),
            )



########################################################################
# File name: service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import collections
import hashlib
import logging
import warnings

import aioxmpp
import aioxmpp.callbacks as callbacks
import aioxmpp.service as service
import aioxmpp.disco as disco
import aioxmpp.pep as pep
import aioxmpp.presence as presence
import aioxmpp.pubsub as pubsub
import aioxmpp.vcard as vcard

from aioxmpp.cache import LRUDict
from aioxmpp.utils import namespaces, gather_reraise_multi

from . import xso as avatar_xso

logger = logging.getLogger(__name__)


def normalize_id(id_):
    """
    Normalize a SHA1 sum encoded as hexadecimal number in ASCII.

    This does nothing but lowercase the string as to enable robust
    comparison.
    """
    return id_.lower()


class AvatarSet:
    """
    A list of sources of an avatar.

    Exactly one of the sources must include image data in the
    ``image/png`` format. The others provide the location of the
    image data as an URL.

    Adding pointer avatar data is not yet supported.

    .. automethod:: add_avatar_image
    """

    def __init__(self):
        self._image_bytes = None
        self._png_id = None
        self._metadata = avatar_xso.Metadata()

    @property
    def image_bytes(self):
        """
        The image data bytes for MIME type ``text/png``.
        """
        return self._image_bytes

    @property
    def metadata(self):
        """
        The :class:`Metadata` XSO corresponding to this avatar set.
        """
        return self._metadata

    @property
    def png_id(self):
        """
        The SHA1 of the ``image/png`` image data.

        This id is always normalized in the sense of :function:`normalize_id`.
        """
        return self._png_id

    def add_avatar_image(self, mime_type, *, id_=None,
                         image_bytes=None, width=None, height=None,
                         url=None, nbytes=None):
        """
        Add a source of the avatar image.

        All sources of an avatar image added to an avatar set must be
        *the same image*, in different formats and sizes.

        :param mime_type: The MIME type of the avatar image.
        :param id_: The SHA1 of the image data.
        :param nbytes: The size of the image data in bytes.
        :param image_bytes: The image data, this must be supplied only
                            in one call.
        :param url: The URL of the avatar image.
        :param height: The height of the image in pixels (optional).
        :param width: The width of the image in pixels (optional).

        `id_` and `nbytes` may be omitted if and only if `image_data`
        is given and `mime_type` is ``image/png``. If they are
        supplied *and* image data is given, they are checked to match
        the image data.

        It is the caller's responsibility to assure that the provided
        links exist and the files have the correct SHA1 sums.
        """

        if mime_type == "image/png":
            if image_bytes is not None:
                if self._image_bytes is not None:
                    raise RuntimeError(
                        "Only one avatar image may be published directly."
                    )

                sha1 = hashlib.sha1()
                sha1.update(image_bytes)
                id_computed = normalize_id(sha1.hexdigest())
                if id_ is not None:
                    id_ = normalize_id(id_)
                    if id_ != id_computed:
                        raise RuntimeError(
                            "The given id does not match the SHA1 of "
                            "the image data."
                        )
                else:
                    id_ = id_computed

                nbytes_computed = len(image_bytes)
                if nbytes is not None:
                    if nbytes != nbytes_computed:
                        raise RuntimeError(
                            "The given length does not match the length "
                            "of the image data."
                        )
                else:
                    nbytes = nbytes_computed

                self._image_bytes = image_bytes
                self._png_id = id_

        if image_bytes is None and url is None:
            raise RuntimeError(
                "Either the image bytes or an url to retrieve the avatar "
                "image must be given."
            )

        if nbytes is None:
            raise RuntimeError(
                "Image data length is not given an not inferable "
                "from the other arguments."
            )

        if id_ is None:
            raise RuntimeError(
                "The SHA1 of the image data is not given an not inferable "
                "from the other arguments."
            )

        if image_bytes is not None and mime_type != "image/png":
            raise RuntimeError(
                "The image bytes can only be given for image/png data."
            )

        self._metadata.info[mime_type].append(
            avatar_xso.Info(
                id_=id_, mime_type=mime_type, nbytes=nbytes,
                width=width, height=height, url=url
            )
        )


class AbstractAvatarDescriptor:
    """
    Description of the properties of and how to retrieve a specific
    avatar.

    The following attribues are available for all instances:

    .. autoattribute:: remote_jid

    .. autoattribute:: id_

    .. autoattribute:: normalized_id

    .. autoattribute:: can_get_image_bytes_via_xmpp

    .. autoattribute:: has_image_data_in_pubsub

    The following attributes may be :data:`None` and are supposed to
    be used as hints for selection of the avatar to download:

    .. autoattribute:: nbytes

    .. autoattribute:: width

    .. autoattribute:: height

    .. autoattribute:: mime_type

    If this attribute is not :data:`None` it is an URL that points to
    the location of the avatar image:

    .. autoattribute:: url

    The image data belonging to the descriptor can be retrieved by the
    following coroutine:

    .. automethod:: get_image_bytes
    """

    def __init__(self, remote_jid, id_, *, mime_type=None,
                 nbytes=None, width=None, height=None, url=None):
        self._remote_jid = remote_jid
        self._mime_type = mime_type
        self._id = id_
        self._nbytes = nbytes
        self._width = width
        self._height = height
        self._url = url

    def __eq__(self, other):
        return (self._remote_jid == other._remote_jid and
                self._mime_type == other._mime_type and
                self._id == other._id and
                self._nbytes == other._nbytes and
                self._width == other._width and
                self._height == other._height and
                self._url == other._url)

    @asyncio.coroutine
    def get_image_bytes(self):
        """
        Try to retrieve the image data corresponding to this avatar
        descriptor.

        :returns: the image contents
        :rtype: :class:`bytes`

        :raises NotImplementedError: if we do not implement the
            capability to retrieve the image data of this type. It is
            guaranteed to not raise :class:`NotImplementedError` if
            :attr:`can_get_image_bytes_via_xmpp` is true.

        :raises RuntimeError: if the image data described by this
            descriptor is not at the specified location.

        :raises aiomxpp.XMPPCancelError: if trying to retrieve the
            image data causes an XMPP error.
        """
        raise NotImplementedError

    @property
    def can_get_image_bytes_via_xmpp(self):
        """
        Return whether :meth:`get_image_bytes` raises
        :class:`NotImplementedError`.
        """
        return False

    @property
    def has_image_data_in_pubsub(self):
        """
        Whether the image can be retrieved from PubSub.

        .. deprecated:: 0.10

           Use :attr:`can_get_image_bytes_via_xmpp` instead.

           As we support vCard based avatars now the name of this is
           misleading.

           This attribute will be removed in aioxmpp 1.0
        """
        warnings.warn(
            "the has_image_data_in_pubsub attribute is deprecated and will be"
            " removed in 1.0",
            DeprecationWarning,
            stacklevel=1
        )
        return self.can_get_image_bytes_via_xmpp

    @property
    def remote_jid(self):
        """
        The remote JID this avatar belongs to.
        """
        return self._remote_jid

    @property
    def url(self):
        """
        The URL where the avatar image data can be found.

        This may be :data:`None` if the avatar is not given as an URL
        of the image data.
        """
        return self._url

    @property
    def width(self):
        """
        The width of the avatar image in pixels.

        This is :data:`None` if this information is not supplied.
        """
        return self._width

    @property
    def height(self):
        """
        The height of the avatar image in pixels.

        This is :data:`None` if this information is not supplied.
        """
        return self._height

    @property
    def nbytes(self):
        """
        The size of the avatar image data in bytes.
        """
        return self._nbytes

    @property
    def id_(self):
        """
        The SHA1 of the image encoded as hexadecimal number in ASCII.

        This is the original value returned from the underlying
        protocol and should be used for any further interaction with
        the underlying protocol.
        """
        return self._id

    @property
    def normalized_id(self):
        """
        The normalized SHA1 of the image data.

        This is supposed to be used for caching and comparison.
        """
        return normalize_id(self._id)

    @property
    def mime_type(self):
        """
        The MIME type of the image data.
        """
        return self._mime_type


class PubsubAvatarDescriptor(AbstractAvatarDescriptor):

    def __init__(self, remote_jid, id_, *, pubsub=None, **kwargs):
        super().__init__(remote_jid, id_, **kwargs)
        self._pubsub = pubsub

    def __eq__(self, other):
        return (isinstance(other, PubsubAvatarDescriptor) and
                super().__eq__(other))

    @property
    def can_get_image_bytes_via_xmpp(self):
        return True

    @asyncio.coroutine
    def get_image_bytes(self):
        image_data = yield from self._pubsub.get_items_by_id(
            self._remote_jid,
            namespaces.xep0084_data,
            [self.id_],
        )
        if not image_data.payload.items:
            raise RuntimeError("Avatar image data is not set.")

        item, = image_data.payload.items
        return item.registered_payload.data


class HttpAvatarDescriptor(AbstractAvatarDescriptor):

    @asyncio.coroutine
    def get_image_bytes(self):
        raise NotImplementedError

    def __eq__(self, other):
        return (isinstance(other, HttpAvatarDescriptor) and
                super().__eq__(other))

class VCardAvatarDescriptor(AbstractAvatarDescriptor):

    def __init__(self, remote_jid, id_, *, vcard=None, image_bytes=None,
                 **kwargs):
        super().__init__(remote_jid, id_, **kwargs)
        self._vcard = vcard
        self._image_bytes = image_bytes

    def __eq__(self, other):
        # NOTE: we explicitely do *not* check for the equality of
        # image bytes: image bytes is a hidden optimization
        return (isinstance(other, VCardAvatarDescriptor) and
                super().__eq__(other))

    @property
    def can_get_image_bytes_via_xmpp(self):
        return True

    @asyncio.coroutine
    def get_image_bytes(self):
        if self._image_bytes is not None:
            return self._image_bytes

        logger.debug("retrieving vCard %s", self._remote_jid)
        vcard = yield from self._vcard.get_vcard(self._remote_jid)
        photo = vcard.get_photo_data()
        if photo is None:
            raise RuntimeError("Avatar image is not set")

        logger.debug("returning vCard avatar %s", self._remote_jid)
        return photo


class AvatarService(service.Service):
    """
    Access and publish User Avatars (:xep:`84`). Fallback to vCard
    based avatars (:xep:`153`) if no PEP avatar is available.

    This service provides an interface for accessing the avatar of other
    entities in the network, getting notifications on avatar changes and
    publishing an avatar for this entity.

    .. versionchanged:: 0.10

       Support for :xep:`vCard-Based Avatars <153>` was added.

    Observing avatars:

    .. note:: :class:`AvatarService` only caches the metadata, not the
              actual image data. This is the job of the caller.

    .. signal:: on_metadata_changed(jid, metadata)

        Fires when avatar metadata changes.

        :param jid: The JID which the avatar belongs to.
        :param metadata: The new metadata descriptors.
        :type metadata: a sequence of
            :class:`~aioxmpp.avatar.service.AbstractAvatarDescriptor`
            instances

    .. automethod:: get_avatar_metadata

    .. automethod:: subscribe

    Publishing avatars:

    .. automethod:: publish_avatar_set

    .. automethod:: disable_avatar

    .. automethod:: wipe_avatar

    Configuration:

    .. autoattribute:: synchronize_vcard

    .. autoattribute:: advertise_vcard

    .. attribute:: avatar_pep

       The PEP descriptor for claiming the avatar metadata namespace.
       The value is a :class:`~aioxmpp.pep.service.RegisteredPEPNode`,
       whose :attr:`~aioxmpp.pep.service.RegisteredPEPNode.notify`
       property can be used to disable or enable the notification
       feature.

    .. autoattribute:: metadata_cache_size
       :annotation: = 200
    """

    ORDER_AFTER = [
        disco.DiscoClient,
        disco.DiscoServer,
        pubsub.PubSubClient,
        pep.PEPClient,
        vcard.VCardService,
        presence.PresenceClient,
        presence.PresenceServer,
    ]

    avatar_pep = pep.register_pep_node(
        namespaces.xep0084_metadata,
        notify=True,
    )

    on_metadata_changed = callbacks.Signal()

    def __init__(self, client, **kwargs):
        super().__init__(client, **kwargs)
        self._has_pep_avatar = set()
        self._metadata_cache = LRUDict()
        self._metadata_cache.maxsize = 200
        self._pubsub = self.dependencies[pubsub.PubSubClient]
        self._pep = self.dependencies[pep.PEPClient]
        self._presence_server = self.dependencies[presence.PresenceServer]
        self._disco = self.dependencies[disco.DiscoClient]
        self._vcard = self.dependencies[vcard.VCardService]
        # we use this lock to prevent race conditions between different
        # calls of the methods by one client.
        # XXX: Other, independent clients may still cause inconsistent
        # data by race conditions, this should be fixed by at least
        # checking for consistent data after an update.
        self._publish_lock = asyncio.Lock()
        self._synchronize_vcard = False
        self._advertise_vcard = True
        self._vcard_resource_interference = set()
        self._vcard_id = None
        self._vcard_rehashing_for = None
        self._vcard_rehash_task = None

    @property
    def metadata_cache_size(self):
        """
        Maximum number of cache entries in the avatar metadata cache.

        This is mostly a measure to prevent malicious peers from
        exhausting memory by spamming vCard based avatar metadata for
        different resources.

        .. versionadded:: 0.10

        """
        return self._metadata_cache.maxsize

    @metadata_cache_size.setter
    def metadata_cache_size(self, value):
        self._metadata_cache.maxsize = value

    @property
    def synchronize_vcard(self):
        """
        Set this property to true to enable publishing the a vCard avatar.

        This property defaults to false. For the setting true to have
        effect, you have to publish your avatar with :meth:`publish_avatar_set`
        or :meth:`disable_avatar` *after* this switch has been set to true.
        """
        return self._synchronize_vcard

    @synchronize_vcard.setter
    def synchronize_vcard(self, value):
        self._synchronize_vcard = bool(value)

    @property
    def advertise_vcard(self):
        """
        Set this property to false to disable advertisement of the vCard
        avatar via presence broadcast.

        Note, that this reduces traffic, since it makes the presence
        stanzas smaller and we no longer have to recalculate the hash,
        this also disables vCard advertisement for all other
        ressources of the bare local jid, by the business rules of
        :xep:`0153`.

        Note that, when enabling this feature again the vCard has to
        be fetched from the server to recalculate the hash.
        """
        return self._advertise_vcard

    @advertise_vcard.setter
    def advertise_vcard(self, value):
        self._advertise_vcard = bool(value)
        if self._advertise_vcard:
            self._vcard_id = None
            self._start_rehash_task()

    @service.depfilter(aioxmpp.stream.StanzaStream,
                       "service_outbound_presence_filter")
    def _attach_vcard_notify_to_presence(self, stanza):
        if self._advertise_vcard:
            if self._vcard_resource_interference:
                # do not advertise the hash if there is resource interference
                stanza.xep0153_x = avatar_xso.VCardTempUpdate()
            else:
                stanza.xep0153_x = avatar_xso.VCardTempUpdate(self._vcard_id)

        return stanza

    def _update_metadata(self, cache_jid, metadata):
        try:
            cached_metadata = self._metadata_cache[cache_jid]
        except KeyError:
            pass
        else:
            if cached_metadata == metadata:
                return

        self._metadata_cache[cache_jid] = metadata
        self.on_metadata_changed(
            cache_jid,
            metadata
        )

    def _handle_notify(self, full_jid, stanza):
        # handle resource interference as per XEP-153 business rules,
        # we go along with this tracking even if vcard advertisement
        # is off
        if (full_jid.bare() == self.client.local_jid.bare() and
                full_jid != self.client.local_jid):
            if stanza.xep0153_x is None:
                self._vcard_resource_interference.add(full_jid)
            else:
                if self._vcard_resource_interference:
                    self._vcard_resource_interference.discard(full_jid)
                    if not self._vcard_resource_interference:
                        self._vcard_id = None

        # otherwise ignore stanzas without xep0153_x payload, or
        # no photo tag.
        if stanza.xep0153_x is None:
            return

        if stanza.xep0153_x.photo is None:
            return

        # special case MUC presence – otherwise the vcard is retrieved
        # for the bare jid
        if stanza.xep0045_muc_user is not None:
            cache_jid = full_jid
        else:
            cache_jid = full_jid.bare()

        if cache_jid not in self._has_pep_avatar:
            metadata = self._cook_vcard_notify(cache_jid, stanza)
            self._update_metadata(cache_jid, metadata)

        # trigger the download of the vCard and calculation of the
        # vCard avatar hash, if some other resource of our bare jid
        # reported a hash distinct from ours!
        # don't do this if there is a non-compliant resource, we don't
        # send the hash in that case anyway
        if (full_jid.bare() == self.client.local_jid.bare() and
                full_jid != self.client.local_jid and
                self._advertise_vcard and
                not self._vcard_resource_interference):
            if (self._vcard_id is None or
                    stanza.xep0153_x.photo.lower() !=
                    self._vcard_id.lower()):

                # do not rehash if we alread have a rehash task that
                # was triggered by an update with the same hash
                if (self._vcard_rehashing_for is None or
                        self._vcard_rehashing_for !=
                            stanza.xep0153_x.photo.lower()):
                    self._vcard_rehashing_for = stanza.xep0153_x.photo.lower()
                    self._start_rehash_task()

    def _start_rehash_task(self):
        if self._vcard_rehash_task is not None:
            self._vcard_rehash_task.cancel()

        self._vcard_id = None
        # as per XEP immediately resend the presence with empty update
        # element, as this is not synchronous it might already contaiin
        # the new hash, but this is okay as well (as it makes the cached
        # presence stanzas coherent as well).
        self._presence_server.resend_presence()

        self._vcard_rehash_task = asyncio.ensure_future(
            self._calculate_vcard_id()
        )

        def set_new_vcard_id(fut):
            self._vcard_rehashing_for = None
            if not fut.cancelled():
                self._vcard_id = fut.result()

        self._vcard_rehash_task.add_done_callback(
            set_new_vcard_id
        )

    @asyncio.coroutine
    def _calculate_vcard_id(self):
        self.logger.debug("updating vcard hash")
        vcard = yield from self._vcard.get_vcard()
        self.logger.debug("got vcard for hash update: %s", vcard)
        photo = vcard.get_photo_data()

        # if no photo is set in the vcard, set an empty <photo> element
        # in the update; according to the spec this means the avatar
        # is disabled
        if photo is None:
            self.logger.debug("no photo in vcard, advertising as such")
            return ""

        sha1 = hashlib.sha1()
        sha1.update(photo)
        new_hash = sha1.hexdigest().lower()
        self.logger.debug("updated hash to %s", new_hash)
        return new_hash

    @service.depsignal(presence.PresenceClient, "on_available")
    def _handle_on_available(self, full_jid, stanza):
        self._handle_notify(full_jid, stanza)

    @service.depsignal(presence.PresenceClient, "on_changed")
    def _handle_on_changed(self, full_jid, stanza):
        self._handle_notify(full_jid, stanza)

    @service.depsignal(presence.PresenceClient, "on_unavailable")
    def _handle_on_unavailable(self, full_jid, stanza):
        if full_jid.bare() == self.client.local_jid.bare():
            if self._vcard_resource_interference:
                self._vcard_resource_interference.discard(full_jid)
                if not self._vcard_resource_interference:
                    self._start_rehash_task()

        # correctly handle MUC avatars
        if stanza.xep0045_muc_user is not None:
            self._metadata_cache.pop(full_jid, None)

    def _cook_vcard_notify(self, jid, stanza):
        result = []
        # note: an empty photo element correctly
        # results in an empty avatar metadata list
        if stanza.xep0153_x.photo:
            result.append(
                VCardAvatarDescriptor(
                    remote_jid=jid,
                    id_=stanza.xep0153_x.photo,
                    mime_type=None,
                    vcard=self._vcard,
                    nbytes=None,
                )
            )
        return result

    def _cook_metadata(self, jid, items):
        def iter_metadata_info_nodes(items):
            for item in items:
                yield from item.registered_payload.iter_info_nodes()

        result = []
        for info_node in iter_metadata_info_nodes(items):
            if info_node.url is not None:
                descriptor = HttpAvatarDescriptor(
                    remote_jid=jid,
                    id_=info_node.id_,
                    mime_type=info_node.mime_type,
                    nbytes=info_node.nbytes,
                    width=info_node.width,
                    height=info_node.height,
                    url=info_node.url,
                )
            else:
                descriptor = PubsubAvatarDescriptor(
                    remote_jid=jid,
                    id_=info_node.id_,
                    mime_type=info_node.mime_type,
                    nbytes=info_node.nbytes,
                    width=info_node.width,
                    height=info_node.height,
                    pubsub=self._pubsub,
                )
            result.append(descriptor)

        return result

    @service.attrsignal(avatar_pep, "on_item_publish")
    def _handle_pubsub_publish(self, jid, node, item, *, message=None):
        # update the metadata cache
        metadata = self._cook_metadata(jid, [item])
        self._has_pep_avatar.add(jid)
        self._update_metadata(jid, metadata)

    @asyncio.coroutine
    def _get_avatar_metadata_vcard(self, jid):
        logger.debug("trying vCard avatar as fallback for %s", jid)
        vcard = yield from self._vcard.get_vcard(jid)
        photo = vcard.get_photo_data()
        mime_type = vcard.get_photo_mime_type()
        if photo is None:
            return []

        logger.debug("success vCard avatar as fallback for %s",
                     jid)
        sha1 = hashlib.sha1()
        sha1.update(photo)
        return [VCardAvatarDescriptor(
            remote_jid=jid,
            id_=sha1.hexdigest(),
            mime_type=mime_type,
            nbytes=len(photo),
            vcard=self._vcard,
            image_bytes=photo,
        )]

    @asyncio.coroutine
    def _get_avatar_metadata_pep(self, jid):
        try:
            metadata_raw = yield from self._pubsub.get_items(
                jid,
                namespaces.xep0084_metadata,
                max_items=1
            )
        except aioxmpp.XMPPCancelError as e:
            # transparently map feature-not-implemented and
            # item-not-found to be equivalent unset avatar
            if e.condition in (
                    aioxmpp.ErrorCondition.FEATURE_NOT_IMPLEMENTED,
                    aioxmpp.ErrorCondition.ITEM_NOT_FOUND):
                return []
            raise

        self._has_pep_avatar.add(jid)
        return self._cook_metadata(jid, metadata_raw.payload.items)

    @asyncio.coroutine
    def get_avatar_metadata(self, jid, *, require_fresh=False,
                            disable_pep=False):
        """
        Retrieve a list of avatar descriptors.

        :param jid: the JID for which to retrieve the avatar metadata.
        :type jid: :class:`aioxmpp.JID`
        :param require_fresh: if true, do not return results from the
            avatar metadata chache, but retrieve them again from the server.
        :type require_fresh: :class:`bool`
        :param disable_pep: if true, do not try to retrieve the avatar
            via pep, only try the vCard fallback. This usually only
            useful when querying avatars via MUC, where the PEP request
            would be invalid (since it would be for a full jid).
        :type disable_pep: :class:`bool`

        :returns: an iterable of avatar descriptors.
        :rtype: a :class:`list` of
            :class:`~aioxmpp.avatar.service.AbstractAvatarDescriptor`
            instances

        Returning an empty list means that the avatar not set.

        We mask a :class:`XMPPCancelError` in the case that it is
        ``feature-not-implemented`` or ``item-not-found`` and return
        an empty list of avatar descriptors, since this is
        semantically equivalent to not having an avatar.

        .. note::

           It is usually an error to get the avatar for a full jid,
           normally, the avatar is set for the bare jid of a user. The
           exception are vCard avatars over MUC, where the IQ requests
           for the vCard may be translated by the MUC server. It is
           recommended to use the `disable_pep` option in that case.
        """

        if require_fresh:
            self._metadata_cache.pop(jid, None)
        else:
            try:
                return self._metadata_cache[jid]
            except KeyError:
                pass

        if disable_pep:
            metadata = []
        else:
            metadata = yield from self._get_avatar_metadata_pep(jid)

        # try the vcard fallback, note: we don't try this
        # if the PEP avatar is disabled!
        if not metadata and jid not in self._has_pep_avatar:
            metadata = yield from self._get_avatar_metadata_vcard(jid)

        # if a notify was fired while we waited for the results, then
        # use the version in the cache, this will mitigate the race
        # condition because if our version is actually newer we will
        # soon get another notify for this version change!
        if jid not in self._metadata_cache:
            self._update_metadata(jid, metadata)
        return self._metadata_cache[jid]

    @asyncio.coroutine
    def subscribe(self, jid):
        """
        Explicitly subscribe to metadata change notifications for `jid`.
        """
        yield from self._pubsub.subscribe(jid, namespaces.xep0084_metadata)

    @aioxmpp.service.depsignal(aioxmpp.stream.StanzaStream,
                               "on_stream_destroyed")
    def handle_stream_destroyed(self, reason):
        self._metadata_cache.clear()
        self._vcard_resource_interference.clear()
        self._has_pep_avatar.clear()

    @asyncio.coroutine
    def publish_avatar_set(self, avatar_set):
        """
        Make `avatar_set` the current avatar of the jid associated with this
        connection.

        If :attr:`synchronize_vcard` is true and PEP is available the
        vCard is only synchronized if the PEP update is successful.

        This means publishing the ``image/png`` avatar data and the
        avatar metadata set in pubsub. The `avatar_set` must be an
        instance of :class:`AvatarSet`. If :attr:`synchronize_vcard` is
        true the avatar is additionally published in the user vCard.
        """
        id_ = avatar_set.png_id

        done = False
        with (yield from self._publish_lock):
            if (yield from self._pep.available()):
                yield from self._pep.publish(
                    namespaces.xep0084_data,
                    avatar_xso.Data(avatar_set.image_bytes),
                    id_=id_
                )

                yield from self._pep.publish(
                    namespaces.xep0084_metadata,
                    avatar_set.metadata,
                    id_=id_
                )
                done = True

            if self._synchronize_vcard:
                my_vcard = yield from self._vcard.get_vcard()
                my_vcard.set_photo_data("image/png",
                                        avatar_set.image_bytes)
                self._vcard_id = avatar_set.png_id
                yield from self._vcard.set_vcard(my_vcard)
                self._presence_server.resend_presence()
                done = True

        if not done:
            raise RuntimeError(
                "failed to publish avatar: no protocol available"
            )

    @asyncio.coroutine
    def _disable_vcard_avatar(self):
        my_vcard = yield from self._vcard.get_vcard()
        my_vcard.clear_photo_data()
        self._vcard_id = ""
        yield from self._vcard.set_vcard(my_vcard)
        self._presence_server.resend_presence()


    @asyncio.coroutine
    def disable_avatar(self):
        """
        Temporarily disable the avatar.

        If :attr:`synchronize_vcard` is true, the vCard avatar is
        disabled (even if disabling the PEP avatar fails).

        This is done by setting the avatar metadata node empty and if
        :attr:`synchronize_vcard` is true, downloading the vCard,
        removing the avatar data and re-uploading the vCard.

        This method does not error if neither protocol is active.

        :raises aioxmpp.errors.GatherError: if an exception is raised
            by the spawned tasks.
        """

        with (yield from self._publish_lock):
            todo = []
            if self._synchronize_vcard:
                todo.append(self._disable_vcard_avatar())

            if (yield from self._pep.available()):
                todo.append(self._pep.publish(
                    namespaces.xep0084_metadata,
                    avatar_xso.Metadata()
                ))

            yield from gather_reraise_multi(*todo, message="disable_avatar")

    @asyncio.coroutine
    def wipe_avatar(self):
        """
        Remove all avatar data stored on the server.

        If :attr:`synchronize_vcard` is true, the vCard avatar is
        disabled even if disabling the PEP avatar fails.

        This is equivalent to :meth:`disable_avatar` for vCard-based
        avatars, but will also remove the data PubSub node for
        PEP avatars.

        This method does not error if neither protocol is active.

        :raises aioxmpp.errors.GatherError: if an exception is raised
            by the spawned tasks.
        """

        @asyncio.coroutine
        def _wipe_pep_avatar():
            yield from self._pep.publish(
                namespaces.xep0084_metadata,
                avatar_xso.Metadata()
            )
            yield from self._pep.publish(
                namespaces.xep0084_data,
                avatar_xso.Data(b'')
            )

        with (yield from self._publish_lock):
            todo = []
            if self._synchronize_vcard:
                todo.append(self._disable_vcard_avatar())

            if (yield from self._pep.available()):
                todo.append(_wipe_pep_avatar())

            yield from gather_reraise_multi(*todo, message="wipe_avatar")



########################################################################
# File name: __init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.ping` --- XMPP Ping (:xep:`199`)
###############################################

XMPP Ping is a ping on the XMPP protocol level. It can be used to detect
connection liveness (although :class:`aioxmpp.stream.StanzaStream` and thus
:class:`aioxmpp.Client` does that for you) and connectivity/availablility of
remote domains.

Service
=======

.. currentmodule:: aioxmpp

.. autoclass:: PingService()

.. currentmodule:: aioxmpp.ping

XSOs
====

Sometimes it is useful to send a ping manually instead of relying on the
:class:`Service`. For this, the :class:`Ping` IQ payload can be used.

.. autoclass:: Ping()

"""

from .service import PingService
from .xso import Ping



########################################################################
# File name: muc_logger.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import configparser
import locale

from datetime import datetime

import aioxmpp.muc
import aioxmpp.structs

from framework import Example, exec_example


class MucLogger(Example):
    def prepare_argparse(self):
        super().prepare_argparse()

        language_name = locale.getlocale()[0]
        if language_name == "C":
            language_name = "en-gb"

        def language_range(s):
            return aioxmpp.structs.LanguageRange.fromstr(
                s.replace("_", "-")
            )

        default_language = language_range(language_name)

        self.argparse.add_argument(
            "--language",
            default=language_range(language_name),
            type=language_range,
            help="Preferred language: if messages are sent with "
            "multiple languages, this is the language shown "
            "(default: {})".format(default_language),
        )

        # this gives a nicer name in argparse errors
        def jid(s):
            return aioxmpp.JID.fromstr(s)

        self.argparse.add_argument(
            "--muc",
            type=jid,
            default=None,
            help="JID of the muc to join"
        )

        self.argparse.add_argument(
            "--nick",
            default=None,
            help="Nick name to use"
        )

    def configure(self):
        super().configure()

        self.language_selectors = [
            self.args.language,
            aioxmpp.structs.LanguageRange.fromstr("*")
        ]

        self.muc_jid = self.args.muc
        if self.muc_jid is None:
            try:
                self.muc_jid = aioxmpp.JID.fromstr(
                    self.config.get("muc_logger", "muc_jid")
                )
            except (configparser.NoSectionError,
                    configparser.NoOptionError):
                self.muc_jid = aioxmpp.JID.fromstr(
                    input("MUC JID> ")
                )

        self.muc_nick = self.args.nick
        if self.muc_nick is None:
            try:
                self.muc_nick = self.config.get("muc_logger", "nick")
            except (configparser.NoSectionError,
                    configparser.NoOptionError):
                self.muc_nick = input("Nickname> ")

    def make_simple_client(self):
        client = super().make_simple_client()
        muc = client.summon(aioxmpp.MUCClient)
        room, self.room_future = muc.join(
            self.muc_jid,
            self.muc_nick
        )

        room.on_message.connect(self._on_message)
        room.on_topic_changed.connect(self._on_topic_changed)
        room.on_enter.connect(self._on_enter)
        room.on_exit.connect(self._on_exit)
        room.on_leave.connect(self._on_leave)
        room.on_join.connect(self._on_join)

        return client

    def _on_message(self, message, member, source, **kwargs):
        print("{} {}: {}".format(
            datetime.utcnow().isoformat(),
            member.nick,
            message.body.lookup(self.language_selectors),
        ))

    def _on_topic_changed(self, member, new_topic, *, muc_nick=None, **kwargs):
        print("{} *** topic set by {}: {}".format(
            datetime.utcnow().isoformat(),
            member.nick if member is not None else muc_nick,
            new_topic.lookup(self.language_selectors),
        ))

    def _on_enter(self, presence, occupant, **kwargs):
        print("{} *** entered room {}".format(
            datetime.utcnow().isoformat(),
            presence.from_.bare(),
        ))

    def _on_exit(self, **kwargs):
        print("{} *** left room".format(
            datetime.utcnow().isoformat(),
        ))

    def _on_join(self, member, **kwargs):
        print("{} *** {} [{}] entered room".format(
            datetime.utcnow().isoformat(),
            member.nick,
            member.direct_jid,
        ))

    def _on_leave(self, member, muc_leave_mode=None, **kwargs):
        print("{} *** {} [{}] left room ({})".format(
            datetime.utcnow().isoformat(),
            member.nick,
            member.direct_jid,
            muc_leave_mode,
        ))

    @asyncio.coroutine
    def run_example(self):
        self.stop_event = self.make_sigint_event()
        yield from super().run_example()

    @asyncio.coroutine
    def run_simple_example(self):
        print("waiting to join room...")
        done, pending = yield from asyncio.wait(
            [
                self.room_future,
                self.stop_event.wait(),
            ],
            return_when=asyncio.FIRST_COMPLETED
        )
        if self.room_future not in done:
            self.room_future.cancel()
            return

        for fut in pending:
            fut.cancel()

        yield from self.stop_event.wait()


if __name__ == "__main__":
    exec_example(MucLogger())



########################################################################
# File name: test_xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import enum
import io
import unittest
import unittest.mock

from datetime import datetime

import aioxmpp.forms as forms
import aioxmpp.muc.xso as muc_xso
import aioxmpp.stanza as stanza
import aioxmpp.structs
import aioxmpp.stringprep
import aioxmpp.utils as utils
import aioxmpp.xml
import aioxmpp.xso as xso


TEST_JID = aioxmpp.structs.JID.fromstr(
    "foo@bar.example/fnord"
)


class TestNamespaces(unittest.TestCase):
    def test_base_namespace(self):
        self.assertEqual(
            utils.namespaces.xep0045_muc,
            "http://jabber.org/protocol/muc"
        )

    def test_user_namespace(self):
        self.assertEqual(
            utils.namespaces.xep0045_muc_user,
            "http://jabber.org/protocol/muc#user"
        )

    def test_admin_namespace(self):
        self.assertEqual(
            utils.namespaces.xep0045_muc_admin,
            "http://jabber.org/protocol/muc#admin"
        )

    def test_owner_namespace(self):
        self.assertEqual(
            utils.namespaces.xep0045_muc_owner,
            "http://jabber.org/protocol/muc#owner"
        )

    def test_direct_invite_namespace(self):
        self.assertEqual(
            utils.namespaces.xep0249_conference,
            "jabber:x:conference"
        )


class TestStatusCode(unittest.TestCase):
    def test_is_int_enum(self):
        self.assertTrue(issubclass(
            muc_xso.StatusCode,
            enum.IntEnum,
        ))

    def test_operates_correctly_with_sets(self):
        self.assertEqual(
            muc_xso.StatusCode.AFFILIATION_CHANGE,
            101,
        )
        self.assertSetEqual(
            {muc_xso.StatusCode.AFFILIATION_CHANGE},
            {101}
        )
        self.assertSetEqual(
            {muc_xso.StatusCode.AFFILIATION_CHANGE},
            {101, muc_xso.StatusCode.AFFILIATION_CHANGE}
        )


class TestHistory(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            muc_xso.History,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            muc_xso.History.TAG,
            (utils.namespaces.xep0045_muc, "history")
        )

    def test_maxchars(self):
        self.assertIsInstance(
            muc_xso.History.maxchars,
            xso.Attr
        )
        self.assertEqual(
            muc_xso.History.maxchars.tag,
            (None, "maxchars")
        )
        self.assertIsInstance(
            muc_xso.History.maxchars.type_,
            xso.Integer
        )
        self.assertIs(
            muc_xso.History.maxchars.default,
            None
        )

    def test_maxstanzas(self):
        self.assertIsInstance(
            muc_xso.History.maxstanzas,
            xso.Attr
        )
        self.assertEqual(
            muc_xso.History.maxstanzas.tag,
            (None, "maxstanzas")
        )
        self.assertIsInstance(
            muc_xso.History.maxstanzas.type_,
            xso.Integer
        )
        self.assertIs(
            muc_xso.History.maxstanzas.default,
            None
        )

    def test_seconds(self):
        self.assertIsInstance(
            muc_xso.History.seconds,
            xso.Attr
        )
        self.assertEqual(
            muc_xso.History.seconds.tag,
            (None, "seconds")
        )
        self.assertIsInstance(
            muc_xso.History.seconds.type_,
            xso.Integer
        )
        self.assertIs(
            muc_xso.History.seconds.default,
            None
        )

    def test_since(self):
        self.assertIsInstance(
            muc_xso.History.since,
            xso.Attr
        )
        self.assertEqual(
            muc_xso.History.since.tag,
            (None, "since")
        )
        self.assertIsInstance(
            muc_xso.History.since.type_,
            xso.DateTime
        )
        self.assertIs(
            muc_xso.History.since.default,
            None
        )

    def test_init(self):
        hist = muc_xso.History()
        self.assertIsNone(hist.seconds)
        self.assertIsNone(hist.since)
        self.assertIsNone(hist.maxstanzas)
        self.assertIsNone(hist.maxchars)

        now = datetime.utcnow()
        hist = muc_xso.History(
            since=now,
            seconds=123,
            maxstanzas=345,
            maxchars=456
        )
        self.assertEqual(hist.seconds, 123)
        self.assertEqual(hist.since, now)
        self.assertEqual(hist.maxchars, 456)
        self.assertEqual(hist.maxstanzas, 345)

        with self.assertRaisesRegex(TypeError, "positional argument"):
            hist = muc_xso.History(123)


class TestGenericExt(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            muc_xso.GenericExt,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            muc_xso.GenericExt.TAG,
            (utils.namespaces.xep0045_muc, "x")
        )

    def test_history(self):
        self.assertIsInstance(
            muc_xso.GenericExt.history,
            xso.Child
        )
        self.assertIn(
            muc_xso.History,
            muc_xso.GenericExt.history._classes
        )
        self.assertFalse(
            muc_xso.GenericExt.history.required
        )

    def test_password(self):
        self.assertIsInstance(
            muc_xso.GenericExt.password,
            xso.ChildText
        )
        self.assertEqual(
            muc_xso.GenericExt.password.tag,
            (utils.namespaces.xep0045_muc, "password")
        )
        self.assertIsNone(
            muc_xso.GenericExt.password.default
        )

    def test_Presence_attr(self):
        self.assertIsInstance(
            stanza.Presence.xep0045_muc,
            xso.Child
        )
        self.assertIn(
            muc_xso.GenericExt,
            stanza.Presence.xep0045_muc._classes
        )
        self.assertFalse(
            stanza.Presence.xep0045_muc.required
        )

    def test_Message_attr(self):
        self.assertIsInstance(
            stanza.Message.xep0045_muc,
            xso.Child
        )
        self.assertIn(
            muc_xso.GenericExt,
            stanza.Message.xep0045_muc._classes
        )
        self.assertFalse(
            stanza.Message.xep0045_muc.required
        )


class TestStatus(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            muc_xso.Status,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            muc_xso.Status.TAG,
            (utils.namespaces.xep0045_muc_user, "status")
        )

    def test_code(self):
        self.assertIsInstance(
            muc_xso.Status.code,
            xso.Attr
        )
        self.assertEqual(
            muc_xso.Status.code.tag,
            (None, "code")
        )
        self.assertIsInstance(
            muc_xso.Status.code.type_,
            xso.EnumCDataType
        )
        self.assertIsInstance(
            muc_xso.Status.code.type_.nested_type,
            xso.Integer,
        )
        self.assertEqual(
            muc_xso.Status.code.type_.enum_class,
            muc_xso.StatusCode
        )
        self.assertTrue(muc_xso.Status.code.type_.allow_coerce)
        self.assertFalse(muc_xso.Status.code.type_.deprecate_coerce)
        self.assertTrue(muc_xso.Status.code.type_.pass_unknown)
        self.assertEqual(
            muc_xso.Status.code.default,
            xso.NO_DEFAULT
        )

    def test_init(self):
        item = muc_xso.Status(muc_xso.StatusCode.CONFIG_SEMI_ANONYMOUS)
        self.assertEqual(item.code, 173)
        self.assertEqual(item.code, muc_xso.StatusCode.CONFIG_SEMI_ANONYMOUS)

    def test_can_deserialise_arbitrary_integer(self):
        src = io.BytesIO(
            b"<status xmlns='http://jabber.org/protocol/muc#user' code='1234'/>"
        )
        item = aioxmpp.xml.read_single_xso(
            src,
            muc_xso.Status,
        )

        self.assertEqual(
            item.code,
            1234,
        )

    def test_prefers_enum_if_possible(self):
        src = io.BytesIO(
            b"<status xmlns='http://jabber.org/protocol/muc#user' code='101'/>"
        )
        item = aioxmpp.xml.read_single_xso(
            src,
            muc_xso.Status,
        )

        self.assertIsInstance(
            item.code,
            muc_xso.StatusCode,
        )

    def test_can_serialise_arbitrary_integer(self):
        buf = io.BytesIO()
        aioxmpp.xml.write_single_xso(
            muc_xso.Status(1234),
            buf,
        )
        buf.seek(0, io.SEEK_SET)

        item = aioxmpp.xml.read_single_xso(
            buf,
            muc_xso.Status,
        )

        self.assertEqual(
            item.code,
            1234,
        )

    def test_coerces_if_possible(self):
        item = muc_xso.Status(201)
        self.assertIsInstance(item.code, muc_xso.StatusCode)
        self.assertEqual(item.code, 201)

    def test_can_serialise_enum_member(self):
        buf = io.BytesIO()
        aioxmpp.xml.write_single_xso(
            muc_xso.Status(muc_xso.StatusCode.CREATED),
            buf,
        )
        buf.seek(0, io.SEEK_SET)

        item = aioxmpp.xml.read_single_xso(
            buf,
            muc_xso.Status,
        )

        self.assertEqual(
            item.code,
            muc_xso.StatusCode.CREATED,
        )


class TestStatusCodeList(unittest.TestCase):
    def test_is_element_type(self):
        self.assertTrue(issubclass(
            muc_xso.StatusCodeList,
            xso.AbstractElementType
        ))

    def setUp(self):
        self.type_ = muc_xso.StatusCodeList()

    def tearDown(self):
        del self.type_

    def test_unpack(self):
        item = muc_xso.Status(123)

        self.assertEqual(
            self.type_.unpack(item),
            123
        )

    def test_pack(self):
        item = self.type_.pack(123)
        self.assertIsInstance(
            item,
            muc_xso.Status
        )
        self.assertEqual(item.code, 123)

    def test_get_xso_types(self):
        self.assertCountEqual(
            self.type_.get_xso_types(),
            [muc_xso.Status]
        )


class TestDestroyNotification(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            muc_xso.DestroyNotification,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            muc_xso.DestroyNotification.TAG,
            (utils.namespaces.xep0045_muc_user, "destroy")
        )

    def test_reason(self):
        self.assertIsInstance(
            muc_xso.DestroyNotification.reason,
            xso.ChildText
        )
        self.assertEqual(
            muc_xso.DestroyNotification.reason.tag,
            (utils.namespaces.xep0045_muc_user, "reason")
        )
        self.assertIsNone(
            muc_xso.DestroyNotification.reason.default
        )

    def test_jid(self):
        self.assertIsInstance(
            muc_xso.DestroyNotification.jid,
            xso.Attr
        )
        self.assertEqual(
            muc_xso.DestroyNotification.jid.tag,
            (None, "jid")
        )
        self.assertIsInstance(
            muc_xso.DestroyNotification.jid.type_,
            xso.JID
        )
        self.assertIsNone(muc_xso.DestroyNotification.jid.default)


class TestDecline(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            muc_xso.Decline,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            muc_xso.Decline.TAG,
            (utils.namespaces.xep0045_muc_user, "decline")
        )

    def test_from_(self):
        self.assertIsInstance(
            muc_xso.Decline.from_,
            xso.Attr
        )
        self.assertEqual(
            muc_xso.Decline.from_.tag,
            (None, "from")
        )
        self.assertIsInstance(
            muc_xso.Decline.from_.type_,
            xso.JID
        )
        self.assertIsNone(muc_xso.Decline.from_.default)

    def test_to(self):
        self.assertIsInstance(
            muc_xso.Decline.to,
            xso.Attr
        )
        self.assertEqual(
            muc_xso.Decline.to.tag,
            (None, "to")
        )
        self.assertIsInstance(
            muc_xso.Decline.to.type_,
            xso.JID
        )
        self.assertIsNone(muc_xso.Decline.to.default)

    def test_reason(self):
        self.assertIsInstance(
            muc_xso.Decline.reason,
            xso.ChildText
        )
        self.assertEqual(
            muc_xso.Decline.reason.tag,
            (utils.namespaces.xep0045_muc_user, "reason")
        )
        self.assertIsNone(
            muc_xso.Decline.reason.default
        )


class TestInvite(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            muc_xso.Invite,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            muc_xso.Invite.TAG,
            (utils.namespaces.xep0045_muc_user, "invite")
        )

    def test_from_(self):
        self.assertIsInstance(
            muc_xso.Invite.from_,
            xso.Attr
        )
        self.assertEqual(
            muc_xso.Invite.from_.tag,
            (None, "from")
        )
        self.assertIsInstance(
            muc_xso.Invite.from_.type_,
            xso.JID
        )
        self.assertIsNone(muc_xso.Invite.from_.default)

    def test_to(self):
        self.assertIsInstance(
            muc_xso.Invite.to,
            xso.Attr
        )
        self.assertEqual(
            muc_xso.Invite.to.tag,
            (None, "to")
        )
        self.assertIsInstance(
            muc_xso.Invite.to.type_,
            xso.JID
        )
        self.assertIsNone(muc_xso.Invite.to.default)

    def test_reason(self):
        self.assertIsInstance(
            muc_xso.Invite.reason,
            xso.ChildText
        )
        self.assertEqual(
            muc_xso.Invite.reason.tag,
            (utils.namespaces.xep0045_muc_user, "reason")
        )
        self.assertIsNone(
            muc_xso.Invite.reason.default
        )

    def test_password(self):
        self.assertIsInstance(
            muc_xso.Invite.password,
            xso.ChildText
        )
        self.assertEqual(
            muc_xso.Invite.password.tag,
            (utils.namespaces.xep0045_muc_user, "password")
        )
        self.assertIsNone(
            muc_xso.Invite.password.default
        )


class TestActorBase(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            muc_xso.ActorBase,
            xso.XSO
        ))

    def test_jid(self):
        self.assertIsInstance(
            muc_xso.ActorBase.jid,
            xso.Attr
        )
        self.assertEqual(
            muc_xso.ActorBase.jid.tag,
            (None, "jid")
        )
        self.assertIsInstance(
            muc_xso.ActorBase.jid.type_,
            xso.JID
        )
        self.assertIsNone(muc_xso.ActorBase.jid.default)

    def test_nick(self):
        self.assertIsInstance(
            muc_xso.ActorBase.nick,
            xso.Attr
        )
        self.assertEqual(
            muc_xso.ActorBase.nick.tag,
            (None, "nick")
        )
        self.assertIsInstance(
            muc_xso.ActorBase.nick.type_,
            xso.String
        )
        self.assertEqual(
            muc_xso.ActorBase.nick.type_.prepfunc,
            aioxmpp.stringprep.resourceprep
        )


class TestItemBase(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            muc_xso.ItemBase,
            xso.XSO
        ))

    def test_affiliation(self):
        self.assertIsInstance(
            muc_xso.ItemBase.affiliation,
            xso.Attr
        )
        self.assertEqual(
            muc_xso.ItemBase.affiliation.tag,
            (None, "affiliation")
        )
        self.assertIsInstance(
            muc_xso.ItemBase.affiliation.validator,
            xso.RestrictToSet
        )
        self.assertSetEqual(
            muc_xso.ItemBase.affiliation.validator.values,
            {
                "admin",
                "member",
                "none",
                "outcast",
                "owner",
                None,
            }
        )
        self.assertEqual(
            muc_xso.ItemBase.affiliation.validate,
            xso.ValidateMode.ALWAYS
        )
        self.assertEqual(
            muc_xso.ItemBase.affiliation.default,
            None
        )

    def test_jid(self):
        self.assertIsInstance(
            muc_xso.ItemBase.jid,
            xso.Attr
        )
        self.assertEqual(
            muc_xso.ItemBase.jid.tag,
            (None, "jid")
        )
        self.assertIsInstance(
            muc_xso.ItemBase.jid.type_,
            xso.JID
        )
        self.assertIsNone(muc_xso.ItemBase.jid.default)

    def test_nick(self):
        self.assertIsInstance(
            muc_xso.ItemBase.nick,
            xso.Attr
        )
        self.assertEqual(
            muc_xso.ItemBase.nick.tag,
            (None, "nick")
        )
        self.assertIsInstance(
            muc_xso.ItemBase.nick.type_,
            xso.String
        )
        self.assertEqual(
            muc_xso.ItemBase.nick.type_.prepfunc,
            aioxmpp.stringprep.resourceprep
        )

    def test_role(self):
        self.assertIsInstance(
            muc_xso.ItemBase.role,
            xso.Attr
        )
        self.assertEqual(
            muc_xso.ItemBase.role.tag,
            (None, "role")
        )
        self.assertIsInstance(
            muc_xso.ItemBase.role.validator,
            xso.RestrictToSet
        )
        self.assertSetEqual(
            muc_xso.ItemBase.role.validator.values,
            {
                "moderator",
                "none",
                "participant",
                "visitor",
                None,
            }
        )
        self.assertEqual(
            muc_xso.ItemBase.role.validate,
            xso.ValidateMode.ALWAYS
        )
        self.assertEqual(
            muc_xso.ItemBase.role.default,
            None
        )


class TestUserActor(unittest.TestCase):
    def test_is_actor_base(self):
        self.assertTrue(issubclass(
            muc_xso.UserActor,
            muc_xso.ActorBase
        ))

    def test_tag(self):
        self.assertEqual(
            muc_xso.UserActor.TAG,
            (utils.namespaces.xep0045_muc_user, "actor")
        )


class TestContinue(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            muc_xso.Continue,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            muc_xso.Continue.TAG,
            (utils.namespaces.xep0045_muc_user, "continue")
        )

    def test_thread(self):
        self.assertIsInstance(
            muc_xso.Continue.thread,
            xso.Attr
        )
        self.assertEqual(
            muc_xso.Continue.thread.tag,
            (None, "thread")
        )
        self.assertEqual(
            muc_xso.Continue.thread.type_,
            stanza.Thread.identifier.type_
        )
        self.assertIsNone(
            muc_xso.Continue.thread.default,
        )


class TestUserItem(unittest.TestCase):
    def test_is_item_base(self):
        self.assertTrue(issubclass(
            muc_xso.UserItem,
            muc_xso.ItemBase
        ))

    def test_tag(self):
        self.assertEqual(
            muc_xso.UserItem.TAG,
            (utils.namespaces.xep0045_muc_user, "item")
        )

    def test_actor(self):
        self.assertIsInstance(
            muc_xso.UserItem.actor,
            xso.Child
        )
        self.assertSetEqual(
            muc_xso.UserItem.actor._classes,
            {muc_xso.UserActor}
        )

    def test_continue_(self):
        self.assertIsInstance(
            muc_xso.UserItem.continue_,
            xso.Child
        )
        self.assertSetEqual(
            muc_xso.UserItem.continue_._classes,
            {muc_xso.Continue}
        )

    def test_reason(self):
        self.assertIsInstance(
            muc_xso.UserItem.reason,
            xso.ChildText
        )
        self.assertEqual(
            muc_xso.UserItem.reason.tag,
            (utils.namespaces.xep0045_muc_user, "reason")
        )
        self.assertIsNone(
            muc_xso.UserItem.reason.default
        )

    def test_init(self):
        item = muc_xso.UserItem()
        self.assertIsNone(item.affiliation)
        self.assertIsNone(item.jid)
        self.assertIsNone(item.nick)
        self.assertIsNone(item.role)
        self.assertIsNone(item.actor)
        self.assertIsNone(item.continue_)
        self.assertIsNone(item.reason)

        item = muc_xso.UserItem(
            affiliation="admin",
            role="moderator",
            jid=TEST_JID,
            nick="foo",
            reason="fnord"
        )

        self.assertEqual(
            item.affiliation,
            "admin"
        )
        self.assertEqual(
            item.jid,
            TEST_JID
        )
        self.assertEqual(
            item.nick,
            "foo"
        )
        self.assertEqual(
            item.role,
            "moderator",
        )
        self.assertEqual(
            item.reason,
            "fnord"
        )


class TestUserExt(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            muc_xso.UserExt,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            muc_xso.UserExt.TAG,
            (utils.namespaces.xep0045_muc_user, "x")
        )

    def test_status_codes(self):
        self.assertIsInstance(
            muc_xso.UserExt.status_codes,
            xso.ChildValueList
        )
        self.assertIs(
            muc_xso.UserExt.status_codes.container_type,
            set
        )
        self.assertIsInstance(
            muc_xso.UserExt.status_codes.type_,
            muc_xso.StatusCodeList
        )

    def test_destroy(self):
        self.assertIsInstance(
            muc_xso.UserExt.destroy,
            xso.Child,
        )
        self.assertSetEqual(
            muc_xso.UserExt.destroy._classes,
            {muc_xso.DestroyNotification}
        )

    def test_decline(self):
        self.assertIsInstance(
            muc_xso.UserExt.decline,
            xso.Child,
        )
        self.assertSetEqual(
            muc_xso.UserExt.decline._classes,
            {muc_xso.Decline}
        )

    def test_invites(self):
        self.assertIsInstance(
            muc_xso.UserExt.invites,
            xso.ChildList,
        )
        self.assertSetEqual(
            muc_xso.UserExt.invites._classes,
            {muc_xso.Invite}
        )

    def test_items(self):
        self.assertIsInstance(
            muc_xso.UserExt.items,
            xso.ChildList,
        )
        self.assertSetEqual(
            muc_xso.UserExt.items._classes,
            {muc_xso.UserItem}
        )

    def test_password(self):
        self.assertIsInstance(
            muc_xso.UserExt.password,
            xso.ChildText,
        )
        self.assertEqual(
            muc_xso.UserExt.password.tag,
            (utils.namespaces.xep0045_muc_user, "password")
        )
        self.assertIsNone(muc_xso.UserExt.password.default)

    def test_Presence_attr(self):
        self.assertIsInstance(
            stanza.Presence.xep0045_muc_user,
            xso.Child
        )
        self.assertIn(
            muc_xso.UserExt,
            stanza.Presence.xep0045_muc_user._classes
        )
        self.assertFalse(
            stanza.Presence.xep0045_muc_user.required
        )

    def test_Message_attr(self):
        self.assertIsInstance(
            stanza.Message.xep0045_muc_user,
            xso.Child
        )
        self.assertIn(
            muc_xso.UserExt,
            stanza.Message.xep0045_muc_user._classes
        )
        self.assertFalse(
            stanza.Message.xep0045_muc_user.required
        )

    def test_init(self):
        user_ext = muc_xso.UserExt()
        self.assertSetEqual(user_ext.status_codes, set())
        self.assertIsNone(user_ext.destroy)
        self.assertIsNone(user_ext.decline)
        self.assertFalse(user_ext.invites)
        self.assertFalse(user_ext.items)
        self.assertIsNone(user_ext.password)

        items = [
            muc_xso.UserItem(),
            muc_xso.UserItem(),
        ]

        invites = [
            muc_xso.Invite(),
            muc_xso.Invite(),
        ]

        destroy = muc_xso.DestroyNotification()

        decline = muc_xso.Decline()

        user_ext = muc_xso.UserExt(
            status_codes=[110, 110, 200],
            destroy=destroy,
            decline=decline,
            invites=invites,
            items=items,
            password="foobar",
        )
        self.assertSetEqual(user_ext.status_codes, {110, 110, 200})
        self.assertIs(user_ext.destroy, destroy)
        self.assertIs(user_ext.decline, decline)
        self.assertSequenceEqual(user_ext.invites, invites)
        self.assertSequenceEqual(user_ext.items, items)
        self.assertEqual(user_ext.password, "foobar")


class TestAdminActor(unittest.TestCase):
    def test_is_actor_base(self):
        self.assertTrue(issubclass(
            muc_xso.AdminActor,
            muc_xso.ActorBase
        ))

    def test_tag(self):
        self.assertEqual(
            muc_xso.AdminActor.TAG,
            (utils.namespaces.xep0045_muc_admin, "actor")
        )


class TestAdminItem(unittest.TestCase):
    def test_is_item_base(self):
        self.assertTrue(issubclass(
            muc_xso.AdminItem,
            muc_xso.ItemBase
        ))

    def test_tag(self):
        self.assertEqual(
            muc_xso.AdminItem.TAG,
            (utils.namespaces.xep0045_muc_admin, "item")
        )

    def test_actor(self):
        self.assertIsInstance(
            muc_xso.AdminItem.actor,
            xso.Child
        )
        self.assertSetEqual(
            muc_xso.AdminItem.actor._classes,
            {muc_xso.AdminActor}
        )

    def test_reason(self):
        self.assertIsInstance(
            muc_xso.AdminItem.reason,
            xso.ChildText
        )
        self.assertEqual(
            muc_xso.AdminItem.reason.tag,
            (utils.namespaces.xep0045_muc_admin, "reason")
        )
        self.assertIsNone(
            muc_xso.AdminItem.reason.default
        )

    def test_init(self):
        item = muc_xso.AdminItem()
        self.assertIsNone(item.affiliation)
        self.assertIsNone(item.jid)
        self.assertIsNone(item.nick)
        self.assertIsNone(item.role)
        self.assertIsNone(item.actor)
        self.assertIsNone(item.reason)

        item = muc_xso.AdminItem(
            affiliation="admin",
            role="moderator",
            jid=TEST_JID,
            nick="foo",
            reason="foobar",
        )

        self.assertEqual(
            item.affiliation,
            "admin"
        )
        self.assertEqual(
            item.jid,
            TEST_JID
        )
        self.assertEqual(
            item.nick,
            "foo"
        )
        self.assertEqual(
            item.role,
            "moderator",
        )
        self.assertEqual(
            item.reason,
            "foobar"
        )


class TestAdminQuery(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            muc_xso.AdminQuery,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            muc_xso.AdminQuery.TAG,
            (utils.namespaces.xep0045_muc_admin, "query")
        )

    def test_items(self):
        self.assertIsInstance(
            muc_xso.AdminQuery.items,
            xso.ChildList
        )
        self.assertSetEqual(
            muc_xso.AdminQuery.items._classes,
            {muc_xso.AdminItem}
        )

    def test_is_iq_payload(self):
        self.assertIn(
            muc_xso.AdminQuery.TAG,
            stanza.IQ.CHILD_MAP
        )
        self.assertIs(
            stanza.IQ.CHILD_MAP[muc_xso.AdminQuery.TAG],
            stanza.IQ.payload.xq_descriptor
        )

    def test_init(self):
        query = muc_xso.AdminQuery()
        self.assertFalse(query.items)

        item = muc_xso.AdminItem()
        query = muc_xso.AdminQuery(
            items=[
                item
            ]
        )

        self.assertEqual(query.items[0], item)

        with self.assertRaisesRegex(TypeError, "positional argument"):
            muc_xso.AdminQuery([])


class TestDestroyRequest(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            muc_xso.DestroyRequest,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            muc_xso.DestroyRequest.TAG,
            (utils.namespaces.xep0045_muc_owner, "destroy")
        )

    def test_reason(self):
        self.assertIsInstance(
            muc_xso.DestroyRequest.reason,
            xso.ChildText
        )
        self.assertEqual(
            muc_xso.DestroyRequest.reason.tag,
            (utils.namespaces.xep0045_muc_owner, "reason")
        )
        self.assertIsNone(
            muc_xso.DestroyRequest.reason.default
        )

    def test_password(self):
        self.assertIsInstance(
            muc_xso.DestroyRequest.password,
            xso.ChildText
        )
        self.assertEqual(
            muc_xso.DestroyRequest.password.tag,
            (utils.namespaces.xep0045_muc_owner, "password")
        )
        self.assertIsNone(
            muc_xso.DestroyRequest.password.default
        )

    def test_jid(self):
        self.assertIsInstance(
            muc_xso.DestroyRequest.jid,
            xso.Attr
        )
        self.assertEqual(
            muc_xso.DestroyRequest.jid.tag,
            (None, "jid")
        )
        self.assertIsInstance(
            muc_xso.DestroyRequest.jid.type_,
            xso.JID
        )
        self.assertIsNone(muc_xso.DestroyRequest.jid.default)


class TestOwnerQuery(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            muc_xso.OwnerQuery,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            muc_xso.OwnerQuery.TAG,
            (utils.namespaces.xep0045_muc_owner, "query")
        )

    def test_destroy(self):
        self.assertIsInstance(
            muc_xso.OwnerQuery.destroy,
            xso.Child
        )
        self.assertSetEqual(
            muc_xso.OwnerQuery.destroy._classes,
            {muc_xso.DestroyRequest}
        )

    def test_form(self):
        self.assertIsInstance(
            muc_xso.OwnerQuery.form,
            xso.Child
        )
        self.assertSetEqual(
            muc_xso.OwnerQuery.form._classes,
            {forms.Data}
        )

    def test_is_iq_payload(self):
        self.assertIn(
            muc_xso.OwnerQuery.TAG,
            stanza.IQ.CHILD_MAP
        )
        self.assertIs(
            stanza.IQ.CHILD_MAP[muc_xso.OwnerQuery.TAG],
            stanza.IQ.payload.xq_descriptor
        )

    def test_init(self):
        oq = muc_xso.OwnerQuery()
        self.assertIsNone(oq.destroy)
        self.assertIsNone(oq.form)

    def test_init_form(self):
        oq = muc_xso.OwnerQuery(form=unittest.mock.sentinel.form)
        self.assertEqual(
            oq.form,
            unittest.mock.sentinel.form,
        )

    def test_init_destroy(self):
        oq = muc_xso.OwnerQuery(destroy=unittest.mock.sentinel.destroy)
        self.assertEqual(
            oq.destroy,
            unittest.mock.sentinel.destroy,
        )


class TestDirectInvite(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            muc_xso.DirectInvite,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            muc_xso.DirectInvite.TAG,
            (utils.namespaces.xep0249_conference, "x")
        )

    def test_jid(self):
        self.assertIsInstance(
            muc_xso.DirectInvite.jid,
            xso.Attr,
        )
        self.assertEqual(
            muc_xso.DirectInvite.jid.tag,
            (None, "jid")
        )
        self.assertIsInstance(
            muc_xso.DirectInvite.jid.type_,
            xso.JID,
        )
        self.assertIs(
            muc_xso.DirectInvite.jid.default,
            xso.NO_DEFAULT,
        )

    def test_password(self):
        self.assertIsInstance(
            muc_xso.DirectInvite.password,
            xso.Attr,
        )
        self.assertEqual(
            muc_xso.DirectInvite.password.tag,
            (None, "password")
        )
        self.assertIs(
            muc_xso.DirectInvite.password.default,
            None
        )

    def test_reason(self):
        self.assertIsInstance(
            muc_xso.DirectInvite.reason,
            xso.Attr,
        )
        self.assertEqual(
            muc_xso.DirectInvite.reason.tag,
            (None, "reason")
        )
        self.assertIs(
            muc_xso.DirectInvite.reason.default,
            None
        )

    def test_continue(self):
        self.assertIsInstance(
            muc_xso.DirectInvite.continue_,
            xso.Attr,
        )
        self.assertEqual(
            muc_xso.DirectInvite.continue_.tag,
            (None, "continue")
        )
        self.assertIsInstance(
            muc_xso.DirectInvite.continue_.type_,
            xso.Bool,
        )
        self.assertIs(
            muc_xso.DirectInvite.continue_.default,
            False,
        )

    def test_thread(self):
        self.assertIsInstance(
            muc_xso.DirectInvite.thread,
            xso.Attr,
        )
        self.assertEqual(
            muc_xso.DirectInvite.thread.tag,
            (None, "thread")
        )
        self.assertIs(
            muc_xso.DirectInvite.thread.default,
            None
        )

    def test_message_attribute(self):
        self.assertIsInstance(
            aioxmpp.Message.xep0249_direct_invite,
            xso.Child,
        )
        self.assertCountEqual(
            aioxmpp.Message.xep0249_direct_invite._classes,
            [muc_xso.DirectInvite],
        )

    def test_init_requires_jid(self):
        with self.assertRaisesRegex(TypeError, "required"):
            muc_xso.DirectInvite()

    def test_init_default(self):
        di = muc_xso.DirectInvite(TEST_JID)
        self.assertEqual(di.jid, TEST_JID)
        self.assertIsNone(di.reason)
        self.assertIsNone(di.password)
        self.assertFalse(di.continue_)
        self.assertIsNone(di.thread)

    def test_init_full(self):
        di = muc_xso.DirectInvite(
            TEST_JID,
            reason="some reason",
            password="the password",
            continue_=True,
            thread="xyz",
        )
        self.assertEqual(di.jid, TEST_JID)
        self.assertEqual(di.reason, "some reason")
        self.assertEqual(di.password, "the password")
        self.assertTrue(di.continue_)
        self.assertEqual(di.thread, "xyz")



########################################################################
# File name: rfc6120.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.rfc6120` -- Stanza payloads for :rfc:`6120` implementation
#########################################################################

This module contains XSO-related classes for implementation of :rfc:`6120`.

.. autoclass:: BindFeature

.. autoclass:: Bind

.. autoclass:: Required

"""

from . import xso, stanza, nonza
from .utils import namespaces

namespaces.rfc6120_bind = "urn:ietf:params:xml:ns:xmpp-bind"


class Required(xso.XSO):
    """
    The XSO used for the :attr:`.BindFeature.required` attribute.
    """

    TAG = (namespaces.rfc6120_bind, "required")


@nonza.StreamFeatures.as_feature_class
class BindFeature(xso.XSO):
    """
    A stream feature for use with :class:`.nonza.StreamFeatures` which
    indicates that the server allows resource binding.

    .. attribute:: required

       This attribute is either an instance of :class:`Required` or
       :data:`None`. The former indicates that the server requires resource
       binding at this point in the stream negotation; :data:`None` indicates
       that it is not required.

       User code should just test the boolean value of this attribute and not
       worry about the actual types involved.

    """
    TAG = (namespaces.rfc6120_bind, "bind")

    required = xso.Child([Required], required=False)


class Bind(xso.XSO):
    """
    The :class:`.IQ` payload for binding to a resource.

    .. attribute:: jid

       The server-supplied :class:`aioxmpp.JID`. This must not be set by 
       client code.

    .. attribute:: resource

       The client-supplied, optional resource. If a client wishes to bind to a
       specific resource, it must tell the server that using this attribute.

    """

    TAG = (namespaces.rfc6120_bind, "bind")

    jid = xso.ChildText(
        (namespaces.rfc6120_bind, "jid"),
        type_=xso.JID(),
        default=None
    )
    resource = xso.ChildText(
        (namespaces.rfc6120_bind, "resource"),
        default=None
    )

    def __init__(self, jid=None, resource=None):
        super().__init__()
        self.jid = jid
        self.resource = resource


stanza.IQ.register_child(stanza.IQ.payload, Bind)



########################################################################
# File name: test_nonza.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest

import aioxmpp
import aioxmpp.errors as errors
import aioxmpp.nonza as nonza
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces


class TestStreamError(unittest.TestCase):
    def test_declare_ns(self):
        self.assertDictEqual(
            nonza.StreamError.DECLARE_NS,
            {}
        )

    def test_condition_obj_attr(self):
        self.assertIsInstance(
            nonza.StreamError.condition_obj,
            xso.Child,
        )
        self.assertCountEqual(
            [
                member.xso_class
                for member in errors.StreamErrorCondition
            ],
            nonza.StreamError.condition_obj._classes,
        )
        self.assertTrue(nonza.StreamError.condition_obj.required)

    def test_initialises_with_undefined_condition(self):
        e = nonza.StreamError()
        self.assertIsInstance(
            e.condition_obj,
            errors.StreamErrorCondition.UNDEFINED_CONDITION.xso_class,
        )

    @unittest.skipIf(aioxmpp.version_info >= (1, 0, 0),
                     "does not apply to this version of aioxmpp")
    def test_init_works_with_tuple_and_warns(self):
        with self.assertWarnsRegex(
                DeprecationWarning,
                r"as of aioxmpp 1\.0, stream error conditions must be members "
                r"of the aioxmpp\.errors\.StreamErrorCondition "
                r"enumeration") as ctx:
            e = nonza.StreamError(
                errors.StreamErrorCondition.INTERNAL_SERVER_ERROR.value
            )

        self.assertEqual(
            e.condition,
            errors.StreamErrorCondition.INTERNAL_SERVER_ERROR,
        )

        self.assertTrue(ctx.filename.endswith("test_nonza.py"))

    def test_condition_reflects_enum_member_of_object_after_init(self):
        e = nonza.StreamError()

        self.assertEqual(
            errors.StreamErrorCondition.UNDEFINED_CONDITION,
            e.condition,
        )

    def test_condition_reflects_enum_member_of_object_after_change(self):
        e = nonza.StreamError()
        e.condition_obj = errors.StreamErrorCondition.HOST_GONE.xso_class()

        self.assertEqual(
            errors.StreamErrorCondition.HOST_GONE,
            e.condition,
        )

    def test_setting_condition_replaces_object(self):
        e = nonza.StreamError()
        e.condition = errors.StreamErrorCondition.UNDEFINED_CONDITION

        self.assertEqual(
            e.condition,
            errors.StreamErrorCondition.UNDEFINED_CONDITION
        )

        self.assertIsInstance(
            e.condition_obj,
            errors.StreamErrorCondition.UNDEFINED_CONDITION.xso_class,
        )

    def test_setting_condition_keeps_object_if_condition_matches(self):
        e = nonza.StreamError()
        old = e.condition_obj
        e.condition = errors.StreamErrorCondition.UNDEFINED_CONDITION
        self.assertIs(e.condition_obj, old)

    @unittest.skipIf(aioxmpp.version_info >= (1, 0, 0),
                     "does not apply to this version of aioxmpp")
    def test_accepts_tuple_instead_of_enum_for_condition_and_warns(self):
        e = nonza.StreamError()
        with self.assertWarnsRegex(
                DeprecationWarning,
                r"as of aioxmpp 1\.0, stream error conditions must be members "
                r"of the aioxmpp\.errors\.StreamErrorCondition "
                r"enumeration") as ctx:
            e.condition = errors.StreamErrorCondition.HOST_GONE.value

        self.assertEqual(
            errors.StreamErrorCondition.HOST_GONE,
            e.condition,
        )

        self.assertIsInstance(
            e.condition_obj,
            errors.StreamErrorCondition.HOST_GONE.xso_class,
        )

        self.assertTrue(ctx.filename.endswith("test_nonza.py"))

    def test_from_exception(self):
        obj = nonza.StreamError.from_exception(errors.StreamError(
            errors.StreamErrorCondition.UNDEFINED_CONDITION,
            text="foobar"
        ))

        self.assertEqual(
            errors.StreamErrorCondition.UNDEFINED_CONDITION,
            obj.condition
        )

        self.assertEqual(
            "foobar",
            obj.text
        )

    def test_to_exception(self):
        obj = nonza.StreamError()
        obj.condition = errors.StreamErrorCondition.RESTRICTED_XML
        obj.text = "foobar"

        exc = obj.to_exception()

        self.assertIsInstance(
            exc,
            errors.StreamError
        )

        self.assertEqual(
            errors.StreamErrorCondition.RESTRICTED_XML,
            exc.condition
        )

        self.assertEqual(
            "foobar",
            exc.text
        )

    def test_default_init(self):
        obj = nonza.StreamError()

        self.assertEqual(
            errors.StreamErrorCondition.UNDEFINED_CONDITION,
            obj.condition
        )

        self.assertIsNone(obj.text)

    def test_init(self):
        obj = nonza.StreamError(
            condition=errors.StreamErrorCondition.RESET,
            text="foobar"
        )

        self.assertEqual(
            errors.StreamErrorCondition.RESET,
            obj.condition
        )

        self.assertEqual(
            "foobar",
            obj.text
        )


class TestStreamFeatures(unittest.TestCase):
    def test_setup(self):
        self.assertEqual(
            (namespaces.xmlstream, "features"),
            nonza.StreamFeatures.TAG
        )
        self.assertDictEqual(
            {
                None: namespaces.xmlstream,
            },
            nonza.StreamFeatures.DECLARE_NS
        )
        self.assertEqual(
            xso.UnknownChildPolicy.DROP,
            nonza.StreamFeatures.UNKNOWN_CHILD_POLICY
        )

    def test_as_feature_class_decorator(self):
        class FakeFeature(xso.XSO):
            TAG = ("uri:foo", "bar")

        self.assertNotIn(
            FakeFeature.TAG,
            nonza.StreamFeatures.CHILD_MAP
        )

        FakeFeature = nonza.StreamFeatures.as_feature_class(
            FakeFeature
        )

        self.assertTrue(issubclass(FakeFeature, xso.XSO))

        self.assertIn(
            FakeFeature.TAG,
            nonza.StreamFeatures.CHILD_MAP
        )
        self.assertEqual(
            nonza.StreamFeatures.features,
            nonza.StreamFeatures.CHILD_MAP[FakeFeature.TAG]
        )

    def test_is_feature(self):
        class FakeFeature(xso.XSO):
            TAG = ("uri:foo", "baz")

        self.assertFalse(nonza.StreamFeatures.is_feature(FakeFeature))

        nonza.StreamFeatures.as_feature_class(FakeFeature)

        self.assertTrue(nonza.StreamFeatures.is_feature(FakeFeature))

    def test__getitem__(self):
        class FakeFeature(xso.XSO):
            TAG = ("uri:foo", "foo")

        class NotAFeature(xso.XSO):
            TAG = ("uri:foo", "bar")

        instance = FakeFeature()

        features = nonza.StreamFeatures()
        features.features[FakeFeature.TAG].append(instance)

        self.assertIs(
            instance,
            features[FakeFeature]
        )

        with self.assertRaises(KeyError) as ctx:
            features[NotAFeature]

        self.assertEqual(
            (NotAFeature,),
            ctx.exception.args
        )
        self.assertIsNone(ctx.exception.__cause__)
        self.assertTrue(ctx.exception.__suppress_context__)

    def test___delitem__(self):
        class FakeFeature(xso.XSO):
            TAG = ("uri:foo", "foo")

        instance = FakeFeature()

        features = nonza.StreamFeatures()
        features.features[FakeFeature.TAG].append(instance)

        del features[FakeFeature]

        with self.assertRaises(KeyError):
            features[FakeFeature]

        with self.assertRaises(KeyError):
            del features[FakeFeature]

    def test___setitem__(self):
        class FakeFeature(xso.XSO):
            TAG = ("uri:foo", "foo")

        class NotAFeature(xso.XSO):
            TAG = ("uri:foo", "foo")  # using the same tag here!

        instance = FakeFeature()

        features = nonza.StreamFeatures()
        features[...] = instance

        self.assertIs(
            instance,
            features[FakeFeature]
        )

        with self.assertRaisesRegex(ValueError, "incorrect XSO class"):
            features[NotAFeature] = instance

        del features[FakeFeature]


    def test_get_feature(self):
        class FakeFeature(xso.XSO):
            TAG = ("uri:foo", "foo")

        class NotAFeature(xso.XSO):
            TAG = ("uri:foo", "bar")

        instance = FakeFeature()

        default = object()

        features = nonza.StreamFeatures()
        features.features[FakeFeature.TAG].append(instance)

        self.assertIs(instance, features.get_feature(FakeFeature))
        self.assertIsNone(features.get_feature(NotAFeature))
        self.assertIs(
            default,
            features.get_feature(NotAFeature, default=default)
        )

    def test_has_feature(self):
        class FakeFeature(xso.XSO):
            TAG = ("uri:foo", "foo")

        class NotAFeature(xso.XSO):
            TAG = ("uri:foo", "bar")

        features = nonza.StreamFeatures()
        features.features[FakeFeature.TAG].append(FakeFeature())

        self.assertTrue(features.has_feature(FakeFeature))
        self.assertFalse(features.has_feature(NotAFeature))

    def test_contains(self):
        features = nonza.StreamFeatures()
        with self.assertRaisesRegex(TypeError,
                                    "membership test not supported"):
            "foo" in features

    def test_iter_features(self):
        class FakeFeatureA(xso.XSO):
            TAG = ("uri:foo", "foo")

        class FakeFeatureB(xso.XSO):
            TAG = ("uri:foo", "bar")

        instance1, instance2 = FakeFeatureA(), FakeFeatureB()

        features = nonza.StreamFeatures()
        features[...] = instance1
        features[...] = instance2

        self.assertSetEqual(
            {
                instance1,
                instance2
            },
            set(features)
        )


class TestStartTLSXSO(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            nonza.StartTLSXSO,
            xso.XSO
        ))

    def test_declare_ns(self):
        self.assertDictEqual(
            nonza.StartTLSXSO.DECLARE_NS,
            {
                None: namespaces.starttls
            }
        )


class TestStartTLSFeature(unittest.TestCase):
    def test_is_starttls_xso(self):
        self.assertTrue(issubclass(
            nonza.StartTLSFeature,
            nonza.StartTLSXSO
        ))

    def test_tag(self):
        self.assertEqual(
            nonza.StartTLSFeature.TAG,
            (namespaces.starttls, "starttls")
        )

    def test_required(self):
        self.assertIsInstance(
            nonza.StartTLSFeature.required,
            xso.Child
        )
        self.assertSetEqual(
            nonza.StartTLSFeature.required._classes,
            {nonza.StartTLSFeature.Required},
        )
        self.assertFalse(nonza.StartTLSFeature.required.required)

    def test_is_registered_stream_feature(self):
        self.assertTrue(nonza.StreamFeatures.is_feature(
            nonza.StartTLSFeature
        ))


class TestStartTLSFeature_Required(unittest.TestCase):
    def test_is_starttls_xso(self):
        self.assertTrue(issubclass(
            nonza.StartTLSFeature.Required,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            nonza.StartTLSFeature.Required.TAG,
            (namespaces.starttls, "required")
        )


class TestStartTLS(unittest.TestCase):
    def test_is_starttls_xso(self):
        self.assertTrue(issubclass(
            nonza.StartTLS,
            nonza.StartTLSXSO
        ))

    def test_tag(self):
        self.assertEqual(
            nonza.StartTLS.TAG,
            (namespaces.starttls, "starttls")
        )


class TestStartTLSFailure(unittest.TestCase):
    def test_is_starttls_xso(self):
        self.assertTrue(issubclass(
            nonza.StartTLSFailure,
            nonza.StartTLSXSO
        ))

    def test_tag(self):
        self.assertEqual(
            nonza.StartTLSFailure.TAG,
            (namespaces.starttls, "failure")
        )


class TestStartTLSProceed(unittest.TestCase):
    def test_is_starttls_xso(self):
        self.assertTrue(issubclass(
            nonza.StartTLSProceed,
            nonza.StartTLSXSO
        ))

    def test_tag(self):
        self.assertEqual(
            nonza.StartTLSProceed.TAG,
            (namespaces.starttls, "proceed")
        )


class TestSASLXSO(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            nonza.SASLXSO,
            xso.XSO
        ))

    def test_declare_ns(self):
        self.assertDictEqual(
            nonza.SASLXSO.DECLARE_NS,
            {
                None: namespaces.sasl
            }
        )


class TestSASLAuth(unittest.TestCase):
    def test_is_sasl_xso(self):
        self.assertTrue(issubclass(
            nonza.SASLAuth,
            nonza.SASLXSO
        ))

    def test_tag(self):
        self.assertEqual(
            nonza.SASLAuth.TAG,
            (namespaces.sasl, "auth")
        )

    def test_mechanism(self):
        self.assertIsInstance(
            nonza.SASLAuth.mechanism,
            xso.Attr
        )
        self.assertIs(nonza.SASLAuth.mechanism.default,
                      xso.NO_DEFAULT)


    def test_payload(self):
        self.assertIsInstance(
            nonza.SASLAuth.payload,
            xso.Text
        )
        self.assertIsInstance(
            nonza.SASLAuth.payload.type_,
            xso.Base64Binary
        )
        self.assertIs(
            nonza.SASLAuth.payload.default,
            None
        )

    def test_init(self):
        auth = nonza.SASLAuth("foo")
        self.assertEqual(auth.mechanism, "foo")
        self.assertEqual(auth.payload, None)

        auth = nonza.SASLAuth(mechanism="foo",
                                    payload=b"foobar")
        self.assertEqual(auth.mechanism, "foo")
        self.assertEqual(auth.payload, b"foobar")


class TestSASLChallenge(unittest.TestCase):
    def test_is_sasl_xso(self):
        self.assertTrue(issubclass(
            nonza.SASLChallenge,
            nonza.SASLXSO
        ))

    def test_tag(self):
        self.assertEqual(
            nonza.SASLChallenge.TAG,
            (namespaces.sasl, "challenge")
        )

    def test_payload(self):
        self.assertIsInstance(
            nonza.SASLChallenge.payload,
            xso.Text
        )
        self.assertIsInstance(
            nonza.SASLChallenge.payload.type_,
            xso.Base64Binary
        )
        self.assertIs(
            nonza.SASLChallenge.payload.default,
            xso.NO_DEFAULT
        )

    def test_init(self):
        challenge = nonza.SASLChallenge(b"foo")
        self.assertEqual(challenge.payload, b"foo")

        challenge = nonza.SASLChallenge(payload=b"foo")
        self.assertEqual(challenge.payload, b"foo")

        with self.assertRaisesRegex(TypeError, "positional argument"):
            nonza.SASLChallenge()


class TestSASLResponse(unittest.TestCase):
    def test_is_sasl_xso(self):
        self.assertTrue(issubclass(
            nonza.SASLResponse,
            nonza.SASLXSO
        ))

    def test_tag(self):
        self.assertEqual(
            nonza.SASLResponse.TAG,
            (namespaces.sasl, "response")
        )

    def test_payload(self):
        self.assertIsInstance(
            nonza.SASLResponse.payload,
            xso.Text
        )
        self.assertIsInstance(
            nonza.SASLResponse.payload.type_,
            xso.Base64Binary
        )
        self.assertIs(
            nonza.SASLResponse.payload.default,
            xso.NO_DEFAULT
        )

    def test_init(self):
        challenge = nonza.SASLResponse(b"foo")
        self.assertEqual(challenge.payload, b"foo")

        challenge = nonza.SASLResponse(payload=b"foo")
        self.assertEqual(challenge.payload, b"foo")

        with self.assertRaisesRegex(TypeError, "positional argument"):
            nonza.SASLResponse()


class TestSASLFailure(unittest.TestCase):
    def test_is_sasl_xso(self):
        self.assertTrue(issubclass(
            nonza.SASLFailure,
            nonza.SASLXSO
        ))

    def test_tag(self):
        self.assertEqual(
            nonza.SASLFailure.TAG,
            (namespaces.sasl, "failure")
        )

    def test_condition(self):
        self.assertIsInstance(
            nonza.SASLFailure.condition,
            xso.ChildTag
        )
        self.assertIsInstance(
            nonza.SASLFailure.condition.validator,
            xso.RestrictToSet
        )
        self.assertSetEqual(
            nonza.SASLFailure.condition.validator.values,
            {
                (namespaces.sasl, key)
                for key in [
                        "aborted",
                        "account-disabled",
                        "credentials-expired",
                        "encryption-required",
                        "incorrect-encoding",
                        "invalid-authzid",
                        "invalid-mechanism",
                        "malformed-request",
                        "mechanism-too-weak",
                        "not-authorized",
                        "temporary-auth-failure",
                ]
            }
        )
        self.assertEqual(
            nonza.SASLFailure.condition.declare_prefix,
            None
        )
        self.assertIs(
            nonza.SASLFailure.condition.default,
            xso.NO_DEFAULT
        )

    def test_text(self):
        self.assertIsInstance(
            nonza.SASLFailure.text,
            xso.ChildText
        )
        self.assertEqual(
            nonza.SASLFailure.text.tag,
            (namespaces.sasl, "text")
        )
        self.assertEqual(
            nonza.SASLFailure.text.attr_policy,
            xso.UnknownAttrPolicy.DROP
        )
        self.assertIs(
            nonza.SASLFailure.text.default,
            None
        )
        self.assertIs(
            nonza.SASLFailure.text.declare_prefix,
            None
        )

    def test_init(self):
        fail = nonza.SASLFailure()
        self.assertEqual(
            fail.condition,
            (namespaces.sasl, "temporary-auth-failure")
        )

        fail = nonza.SASLFailure(
            condition=(namespaces.sasl, "invalid-authzid")
        )
        self.assertEqual(
            fail.condition,
            (namespaces.sasl, "invalid-authzid")
        )


class TestSASLSuccess(unittest.TestCase):
    def test_is_sasl_xso(self):
        self.assertTrue(issubclass(
            nonza.SASLSuccess,
            nonza.SASLXSO
        ))

    def test_tag(self):
        self.assertEqual(
            nonza.SASLSuccess.TAG,
            (namespaces.sasl, "success")
        )

    def test_payload(self):
        self.assertIsInstance(
            nonza.SASLSuccess.payload,
            xso.Text
        )
        self.assertIsInstance(
            nonza.SASLSuccess.payload.type_,
            xso.Base64Binary
        )
        self.assertIs(
            nonza.SASLSuccess.payload.default,
            None
        )


class TestSASLAbort(unittest.TestCase):
    def test_is_sasl_xso(self):
        self.assertTrue(issubclass(
            nonza.SASLAbort,
            nonza.SASLXSO
        ))

    def test_tag(self):
        self.assertEqual(
            nonza.SASLAbort.TAG,
            (namespaces.sasl, "abort")
        )



class TestSMXSO(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            nonza.SMXSO,
            xso.XSO
        ))

    def test_declare_ns(self):
        self.assertDictEqual(
            nonza.SMXSO.DECLARE_NS,
            {
                None: namespaces.stream_management
            }
        )


class TestStreamManagementFeature(unittest.TestCase):
    def test_is_sm_xso(self):
        self.assertTrue(issubclass(
            nonza.StreamManagementFeature,
            nonza.SMXSO
        ))

    def test_tag(self):
        self.assertEqual(
            nonza.StreamManagementFeature.TAG,
            (namespaces.stream_management, "sm")
        )

    def test_required(self):
        self.assertIsInstance(
            nonza.StreamManagementFeature.required,
            xso.Child
        )
        self.assertSetEqual(
            nonza.StreamManagementFeature.required._classes,
            {nonza.StreamManagementFeature.Required},
        )
        self.assertFalse(nonza.StreamManagementFeature.required.required)

    def test_optional(self):
        self.assertIsInstance(
            nonza.StreamManagementFeature.optional,
            xso.Child
        )
        self.assertSetEqual(
            nonza.StreamManagementFeature.optional._classes,
            {nonza.StreamManagementFeature.Optional},
        )
        self.assertFalse(nonza.StreamManagementFeature.optional.required)

    def test_is_registered_stream_feature(self):
        self.assertTrue(nonza.StreamFeatures.is_feature(
            nonza.StreamManagementFeature
        ))


class TestStreamManagementFeature_Required(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            nonza.StreamManagementFeature.Required,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            nonza.StreamManagementFeature.Required.TAG,
            (namespaces.stream_management, "required")
        )


class TestStreamManagementFeature_Optional(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            nonza.StreamManagementFeature.Optional,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            nonza.StreamManagementFeature.Optional.TAG,
            (namespaces.stream_management, "optional")
        )


class TestSMRequest(unittest.TestCase):
    def test_is_sm_xso(self):
        self.assertTrue(issubclass(
            nonza.SMRequest,
            nonza.SMXSO
        ))

    def test_tag(self):
        self.assertEqual(
            nonza.SMRequest.TAG,
            (namespaces.stream_management, "r")
        )


class TestSMAcknowledgement(unittest.TestCase):
    def test_is_sm_xso(self):
        self.assertTrue(issubclass(
            nonza.SMAcknowledgement,
            nonza.SMXSO
        ))

    def test_tag(self):
        self.assertEqual(
            nonza.SMAcknowledgement.TAG,
            (namespaces.stream_management, "a")
        )

    def test_default_init(self):
        obj = nonza.SMAcknowledgement()
        self.assertEqual(0, obj.counter)

    def test_init(self):
        obj = nonza.SMAcknowledgement(counter=1234)
        self.assertEqual(1234, obj.counter)


class TestSMEnable(unittest.TestCase):
    def test_is_sm_xso(self):
        self.assertTrue(issubclass(
            nonza.SMEnable,
            nonza.SMXSO
        ))

    def test_tag(self):
        self.assertEqual(
            nonza.SMEnable.TAG,
            (namespaces.stream_management, "enable")
        )

    def test_default_init(self):
        obj = nonza.SMEnable()
        self.assertFalse(obj.resume)
        self.assertIsNone(obj.max_)

    def test_init(self):
        obj = nonza.SMEnable(resume=True, max_=123)
        self.assertTrue(obj.resume)
        self.assertEqual(
            obj.max_,
            123,
        )

    def test_resume(self):
        self.assertIsInstance(
            nonza.SMEnable.resume,
            xso.Attr
        )
        self.assertEqual(
            (None, "resume"),
            nonza.SMEnable.resume.tag
        )
        self.assertIsInstance(
            nonza.SMEnable.resume.type_,
            xso.Bool
        )
        self.assertIs(
            nonza.SMEnable.resume.default,
            False
        )

    def test_max_(self):
        self.assertIsInstance(
            nonza.SMEnable.max_,
            xso.Attr
        )
        self.assertEqual(
            (None, "max"),
            nonza.SMEnable.max_.tag
        )
        self.assertIsInstance(
            nonza.SMEnable.max_.type_,
            xso.Integer
        )
        self.assertIs(
            nonza.SMEnable.max_.default,
            None
        )
        self.assertEqual(
            nonza.SMEnable.max_.validate,
            xso.ValidateMode.ALWAYS,
        )
        self.assertIsInstance(
            nonza.SMEnable.max_.validator,
            xso.NumericRange,
        )
        self.assertEqual(
            nonza.SMEnable.max_.validator.min_,
            0
        )
        self.assertEqual(
            nonza.SMEnable.max_.validator.max_,
            None,
        )


class TestSMEnabled(unittest.TestCase):
    def test_is_sm_xso(self):
        self.assertTrue(issubclass(
            nonza.SMEnabled,
            nonza.SMXSO
        ))

    def test_tag(self):
        self.assertEqual(
            nonza.SMEnabled.TAG,
            (namespaces.stream_management, "enabled")
        )

    def test_default_init(self):
        obj = nonza.SMEnabled()
        self.assertFalse(obj.resume)
        self.assertIsNone(obj.id_)
        self.assertIsNone(obj.location)
        self.assertIsNone(obj.max_)

    def test_init(self):
        obj = nonza.SMEnabled(
            resume=True,
            id_="foobar",
            location=("bar", 1234),
            max_=1023)
        self.assertTrue(obj.resume)
        self.assertEqual("foobar", obj.id_)
        self.assertEqual(("bar", 1234), obj.location)
        self.assertEqual(1023, obj.max_)

    def test_id(self):
        self.assertIsInstance(
            nonza.SMEnabled.id_,
            xso.Attr
        )
        self.assertEqual(
            (None, "id"),
            nonza.SMEnabled.id_.tag
        )
        self.assertIs(
            nonza.SMEnabled.id_.default,
            None,
        )

    def test_location(self):
        self.assertIsInstance(
            nonza.SMEnabled.location,
            xso.Attr
        )
        self.assertEqual(
            (None, "location"),
            nonza.SMEnabled.location.tag
        )
        self.assertIsInstance(
            nonza.SMEnabled.location.type_,
            xso.ConnectionLocation
        )
        self.assertIs(
            nonza.SMEnabled.location.default,
            None
        )

    def test_max(self):
        self.assertIsInstance(
            nonza.SMEnabled.max_,
            xso.Attr
        )
        self.assertEqual(
            (None, "max"),
            nonza.SMEnabled.max_.tag
        )
        self.assertIsInstance(
            nonza.SMEnabled.max_.type_,
            xso.Integer
        )
        self.assertIs(
            nonza.SMEnabled.max_.default,
            None
        )
        self.assertEqual(
            nonza.SMEnabled.max_.validate,
            xso.ValidateMode.ALWAYS,
        )
        self.assertIsInstance(
            nonza.SMEnabled.max_.validator,
            xso.NumericRange,
        )
        self.assertEqual(
            nonza.SMEnabled.max_.validator.min_,
            0
        )
        self.assertEqual(
            nonza.SMEnabled.max_.validator.max_,
            None,
        )

    def test_resume(self):
        self.assertIsInstance(
            nonza.SMEnabled.resume,
            xso.Attr
        )
        self.assertEqual(
            (None, "resume"),
            nonza.SMEnabled.resume.tag
        )
        self.assertIsInstance(
            nonza.SMEnabled.resume.type_,
            xso.Bool
        )
        self.assertIs(
            nonza.SMEnabled.resume.default,
            False
        )


class TestSMResume(unittest.TestCase):
    def test_is_sm_xso(self):
        self.assertTrue(issubclass(
            nonza.SMResume,
            nonza.SMXSO
        ))

    def test_tag(self):
        self.assertEqual(
            nonza.SMResume.TAG,
            (namespaces.stream_management, "resume")
        )

    def test_default_init_not_possible(self):
        with self.assertRaises(TypeError):
            nonza.SMResume()

    def test_init(self):
        obj = nonza.SMResume(
            counter=1,
            previd="foobar")
        self.assertEqual(1, obj.counter)
        self.assertEqual("foobar", obj.previd)

    def test_resume(self):
        self.assertIsInstance(
            nonza.SMResume.counter,
            xso.Attr
        )
        self.assertEqual(
            (None, "h"),
            nonza.SMResume.counter.tag
        )
        self.assertIsInstance(
            nonza.SMResume.counter.type_,
            xso.Integer
        )

    def test_previd(self):
        self.assertIsInstance(
            nonza.SMResume.previd,
            xso.Attr
        )
        self.assertEqual(
            (None, "previd"),
            nonza.SMResume.previd.tag
        )


class TestSMResumed(unittest.TestCase):
    def test_is_sm_xso(self):
        self.assertTrue(issubclass(
            nonza.SMResumed,
            nonza.SMXSO
        ))

    def test_tag(self):
        self.assertEqual(
            nonza.SMResumed.TAG,
            (namespaces.stream_management, "resumed")
        )

    def test_default_init(self):
        with self.assertRaises(TypeError):
            nonza.SMResumed()

    def test_init(self):
        obj = nonza.SMResumed(
            counter=1,
            previd="foobar")
        self.assertEqual(1, obj.counter)
        self.assertEqual("foobar", obj.previd)

    def test_resume(self):
        self.assertIsInstance(
            nonza.SMResumed.counter,
            xso.Attr
        )
        self.assertEqual(
            (None, "h"),
            nonza.SMResumed.counter.tag
        )
        self.assertIsInstance(
            nonza.SMResumed.counter.type_,
            xso.Integer
        )

    def test_previd(self):
        self.assertIsInstance(
            nonza.SMResumed.previd,
            xso.Attr
        )
        self.assertEqual(
            (None, "previd"),
            nonza.SMResumed.previd.tag
        )


class TestSMFailed(unittest.TestCase):
    def test_is_sm_xso(self):
        self.assertTrue(issubclass(
            nonza.SMFailed,
            nonza.SMXSO
        ))

    def test_tag(self):
        self.assertEqual(
            nonza.SMFailed.TAG,
            (namespaces.stream_management, "failed")
        )

    def test_condition(self):
        self.assertIsInstance(
            nonza.SMFailed.condition,
            xso.Child,
        )
        self.assertCountEqual(
            [member.xso_class for member in errors.ErrorCondition],
            nonza.SMFailed.condition._classes,
        )
        self.assertTrue(nonza.SMFailed.condition.required)

    def test_default_init(self):
        obj = nonza.SMFailed()
        self.assertEqual(
            errors.ErrorCondition.UNDEFINED_CONDITION,
            obj.condition.enum_member
        )

    def test_init(self):
        obj = nonza.SMFailed(
            condition=errors.ErrorCondition.ITEM_NOT_FOUND
        )
        self.assertEqual(
            errors.ErrorCondition.ITEM_NOT_FOUND,
            obj.condition.enum_member
        )



########################################################################
# File name: test_hashes.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import hashlib
import unittest
import unittest.mock

import aioxmpp.hashes as hashes
import aioxmpp.xso

from aioxmpp.utils import namespaces


class TestNamespaces(unittest.TestCase):
    def test_namespace(self):
        self.assertEqual(
            namespaces.xep0300_hashes2,
            "urn:xmpp:hashes:2",
        )


class TestHash(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            hashes.Hash,
            aioxmpp.xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            hashes.Hash.TAG,
            (namespaces.xep0300_hashes2, "hash"),
        )

    def test_init_default(self):
        with self.assertRaises(TypeError):
            hashes.Hash()

    def test_init(self):
        h = hashes.Hash("algo", b"digest")
        self.assertEqual(h.algo, "algo")
        self.assertEqual(h.digest, b"digest")

    def test_get_impl_uses_hash_from_algo(self):
        h = hashes.Hash("some algo", b"foo")
        with unittest.mock.patch(
                "aioxmpp.hashes.hash_from_algo") as hash_from_algo:
            impl = h.get_impl()
            hash_from_algo.assert_called_once_with(
                h.algo,
            )
            self.assertEqual(impl, hash_from_algo())


class TestHashType(unittest.TestCase):
    def test_is_element_type(self):
        self.assertTrue(issubclass(
            hashes.HashType,
            aioxmpp.xso.AbstractElementType,
        ))

    def test_get_xso_types(self):
        self.assertCountEqual(
            hashes.HashType.get_xso_types(),
            [hashes.Hash],
        )

    def test_pack(self):
        t = hashes.HashType()
        h = t.pack(("sha-1", b"foobar"))
        self.assertIsInstance(h, hashes.Hash)
        self.assertEqual(h.algo, "sha-1")
        self.assertEqual(h.digest, b"foobar")

    def test_unpack(self):
        t = hashes.HashType()
        h = hashes.Hash("fnord", b"baz")
        pair = t.unpack(h)
        self.assertSequenceEqual(
            pair,
            (
                "fnord",
                b"baz",
            )
        )


class TestHashesParent(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            hashes.HashesParent,
            aioxmpp.xso.XSO,
        ))

    def test_has_no_tag(self):
        self.assertFalse(
            hasattr(hashes.HashesParent, "TAG")
        )

    def test_digests(self):
        self.assertIsInstance(
            hashes.HashesParent.digests,
            aioxmpp.xso.ChildValueMap
        )
        self.assertIsInstance(
            hashes.HashesParent.digests.type_,
            hashes.HashType,
        )


class Testhash_from_algo(unittest.TestCase):
    def test_all_supported(self):
        for algo_name, (enabled,
                        (fun_name,
                         fun_args,
                         fun_kwargs)) in hashes._HASH_ALGO_MAPPING:
            if not enabled:
                continue

            with unittest.mock.patch(
                    "hashlib.{}".format(fun_name),
                    create=True) as hash_impl:
                result = hashes.hash_from_algo(algo_name)

            hash_impl.assert_called_once_with(
                *fun_args,
                **fun_kwargs
            )

            self.assertEqual(
                result,
                hash_impl(),
            )

    def test_raise_ValueError_for_MUST_NOT_hashes(self):
        with self.assertRaisesRegexp(
                ValueError,
                "support of md2 in XMPP is forbidden"):
            hashes.hash_from_algo("md2")

        with self.assertRaisesRegexp(
                ValueError,
                "support of md4 in XMPP is forbidden"):
            hashes.hash_from_algo("md4")

        with self.assertRaisesRegexp(
                ValueError,
                "support of md5 in XMPP is forbidden"):
            hashes.hash_from_algo("md5")

    def test_raises_NotImplementedError_if_function_not_supported(self):
        _sha1 = hashlib.sha1
        with self.assertRaisesRegexp(
                NotImplementedError,
                "sha-1 not supported by hashlib") as ctx:
            del hashlib.sha1
            try:
                hashes.hash_from_algo("sha-1")
            finally:
                hashlib.sha1 = _sha1

        self.assertIsInstance(ctx.exception.__cause__, AttributeError)

    def test_raises_NotImplementedError_if_function_not_defined(self):
        with self.assertRaisesRegexp(
                NotImplementedError,
                "hash algorithm 'foobar' unknown"):
            hashes.hash_from_algo("foobar")


class Testis_algo_supported(unittest.TestCase):
    def test_all_supported(self):
        for algo_name, (enabled,
                        (fun_name,
                         fun_args,
                         fun_kwargs)) in hashes._HASH_ALGO_MAPPING:
            if not enabled:
                self.assertFalse(
                    hashes.is_algo_supported(algo_name)
                )
            else:
                with unittest.mock.patch(
                        "hashlib.{}".format(fun_name),
                        create=True) as hash_impl:
                    self.assertTrue(hashes.is_algo_supported(algo_name))

                hash_impl.assert_not_called()

    def test_return_false_for_MUST_NOT_hashes(self):
        self.assertFalse(hashes.is_algo_supported("md2"))
        self.assertFalse(hashes.is_algo_supported("md4"))
        self.assertFalse(hashes.is_algo_supported("md5"))

    def test_return_false_if_function_not_implemented(self):
        try:
            _sha3_256 = hashlib.sha3_256
        except AttributeError:
            self.assertFalse(hashes.is_algo_supported("sha3-256"))
            return

        self.assertTrue(hashes.is_algo_supported("sha3-256"))
        del hashlib.sha3_256

        try:
            self.assertFalse(hashes.is_algo_supported("sha3-256"))
        finally:
            hashlib.sha3_256 = _sha3_256

    def test_return_false_if_function_not_defined(self):
        self.assertFalse(hashes.is_algo_supported("foobar"))


class Testalgo_from_hashlib(unittest.TestCase):
    def test_all_supported(self):
        for algo_name, (enabled,
                        (fun_name,
                         fun_args,
                         fun_kwargs)) in hashes._HASH_ALGO_MAPPING:
            if not enabled:
                continue

            try:
                impl = hashes.hash_from_algo(algo_name)
            except NotImplementedError:
                continue

            self.assertEqual(
                algo_name,
                hashes.algo_of_hash(impl)
            )

    def test_raise_ValueError_for_MUST_NOT_hashes(self):
        with self.assertRaisesRegexp(
                ValueError,
                "support of md5 in XMPP is forbidden"):
            hashes.algo_of_hash(hashlib.md5())

    def test_sha1(self):
        m = unittest.mock.Mock()
        m.name = "sha1"

        self.assertEqual(
            "sha-1",
            hashes.algo_of_hash(m),
        )

    def test_sha2(self):
        for size in ["224", "256", "384", "512"]:
            m = unittest.mock.Mock()
            m.name = "sha{}".format(size)

            self.assertEqual(
                "sha-{}".format(size),
                hashes.algo_of_hash(m),
            )

    def test_sha3(self):
        for size in ["256", "512"]:
            m = unittest.mock.Mock()
            m.name = "sha3_{}".format(size)

            self.assertEqual(
                "sha3-{}".format(size),
                hashes.algo_of_hash(m),
            )

    def test_blake2b(self):
        versions = [
            (32, "blake2b-256"),
            (64, "blake2b-512"),
        ]

        for digest_size, algo in versions:
            m = unittest.mock.Mock()
            m.digest_size = digest_size
            m.name = "blake2b"

            self.assertEqual(
                algo,
                hashes.algo_of_hash(m),
            )

    def test_raise_ValueError_on_unknown(self):
        m = unittest.mock.Mock()
        with self.assertRaisesRegexp(
                ValueError,
                "unknown hash implementation: <Mock id=.+>"):
            hashes.algo_of_hash(m)


class Testdefault_hash_algorithms(unittest.TestCase):
    def test_selection(self):
        selected = set(hashes.default_hash_algorithms)
        self.assertIn(
            "sha-256",
            selected,
        )

        try:
            hashes.hash_from_algo("sha3-256")
        except NotImplementedError:
            self.assertNotIn("sha3-256", selected)
        else:
            self.assertIn("sha3-256", selected)

        try:
            hashes.hash_from_algo("blake2b-256")
        except NotImplementedError:
            self.assertNotIn("blake2b-256", selected)
        else:
            self.assertIn("blake2b-256", selected)

    def test_all_selected_can_be_instantiated(self):
        for algo in hashes.default_hash_algorithms:
            hashes.hash_from_algo(algo)
